# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Module for listing Digital RF/Metadata files in a directory."""
import bisect
import datetime
import os
import re
import shutil
from collections import defaultdict

import pytz

from . import util

__all__ = (
    'GLOB_DMDFILE', 'GLOB_DMDPROPFILE', 'GLOB_DRFFILE', 'GLOB_DRFPROPFILE',
    'GLOB_SUBDIR', 'RE_DMD', 'RE_DMDPROP', 'RE_DRF', 'RE_DRFDMD',
    'RE_DRFDMDPROP', 'RE_DRFPROP',
    'ilsdrf', 'lsdrf', 'sortkey_drf'
)

# timestamped subdirectory, e.g. 2016-09-21T20-00-00
GLOB_SUBDIR = (
    '[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]'
    'T[0-9][0-9]-[0-9][0-9]-[0-9][0-9]'
)
RE_SUBDIR = (
    r'(?P<year>[0-9]{4})-(?P<month>[0-9]{2})-(?P<day>[0-9]{2})'
    r'T(?P<hour>[0-9]{2})-(?P<minute>[0-9]{2})-(?P<second>[0-9]{2})'
)
_RE_SUBDIR = re.compile('^' + RE_SUBDIR + '$')

# Digital RF file, e.g. rf@1474491360.000.h5
GLOB_DRFFILE = '*@[0-9]*.[0-9][0-9][0-9].h5'
RE_DRFFILE = (
    r'(?P<name>(?!tmp\.).+?)@(?P<secs>[0-9]+)\.(?P<frac>[0-9]{3})\.h5$'
)
_RE_DRFFILE = re.compile('^' + RE_DRFFILE)

# Digital Metadata file, e.g. metadata@1474491360.h5
GLOB_DMDFILE = '*@[0-9]*.h5'
RE_DMDFILE = r'(?P<name>(?!tmp\.).+?)@(?P<secs>[0-9]+)\.h5$'
_RE_DMDFILE = re.compile('^' + RE_DMDFILE)

# either Digital RF or Digital Metadata file
RE_FILE = (
    r'(?P<name>(?!tmp\.).+?)@(?P<secs>[0-9]+)(?:\.(?P<frac>[0-9]{3}))?\.h5$'
)
_RE_FILE = re.compile('^' + RE_FILE)

# Digital RF channel properties file, e.g. drf_properties.h5
# include pre-2.5-style metadata.h5 for now
GLOB_DRFPROPFILE = '*.h5'
RE_DRFPROPFILE = r'(?P<name>drf_properties|metadata)\.h5$'
_RE_DRFPROPFILE = re.compile('^' + RE_DRFPROPFILE)

# Digital RF channel properties file, e.g. dmd_properties.h5
# include pre-2.5-style metadata.h5 for now
GLOB_DMDPROPFILE = '*.h5'
RE_DMDPROPFILE = r'(?P<name>dmd_properties|metadata)\.h5$'
_RE_DMDPROPFILE = re.compile('^' + RE_DMDPROPFILE)

# either Digital RF or Digital Metadata channel properties file
RE_PROPFILE = r'(?P<name>(?:drf|dmd)_properties|metadata)\.h5$'
_RE_PROPFILE = re.compile('^' + RE_PROPFILE)

# Digital RF file in correct subdirectory structure
RE_DRF = re.escape(os.sep).join((r'(?P<chpath>.*?)', RE_SUBDIR, RE_DRFFILE))
# Digital Metadata file in correct subdirectory structure
RE_DMD = re.escape(os.sep).join((r'(?P<chpath>.*?)', RE_SUBDIR, RE_DMDFILE))
# either Digital RF or Digital Metadata in its correct subdirectory
RE_DRFDMD = re.escape(os.sep).join((r'(?P<chpath>.*?)', RE_SUBDIR, RE_FILE))
# properties file associated with Digital RF directory
RE_DRFPROP = re.escape(os.sep).join((r'(?P<chpath>.*?)', RE_DRFPROPFILE))
# properties file associated with Digital Metadata directory
RE_DMDPROP = re.escape(os.sep).join((r'(?P<chpath>.*?)', RE_DMDPROPFILE))
# properties file associated with Digital RF or Digital Metadata directory
RE_DRFDMDPROP = re.escape(os.sep).join((r'(?P<chpath>.*?)', RE_PROPFILE))


def sortkey_drf(filename, regexes=[_RE_FILE]):
    """Get key for a Digital RF filename to sort first by sample time."""
    for r in regexes:
        m = r.match(filename)
        if m:
            try:
                secs = int(m.group('secs'))
            except (IndexError, TypeError):
                # regex matched but there is no 'secs' in regex
                secs = 0
            try:
                frac = int(m.group('frac'))
            except (IndexError, TypeError):
                # regex matched but there is no 'frac' in regex
                frac = 0
            try:
                name = m.group('name')
            except (IndexError, TypeError):
                # regex matched but there is no 'name' in regex
                name = filename
            return (secs*1000 + frac, name)
    return None


def _decorate_drf_files(subdir, filenames, file_regex):
    """Match Digital RF/Metadata filenames and decorate into (time, f)."""
    dec_files = []
    for filename in filenames:
        m = file_regex.match(filename)
        if m:
            try:
                frac = int(m.group('frac'))
            except (IndexError, TypeError):
                # regex matched but there is no 'frac' in regex (metadata)
                frac = 0
            time = datetime.timedelta(
                seconds=int(m.group('secs')),
                milliseconds=frac,
            )
            dec_files.append(
                (time, os.path.join(subdir, filename))
            )
    return dec_files


def _decorated_list_slice(dec_list, starttime=None, endtime=None, ffill=False):
    """Get slice for sorted list of tuples with (time, ...)."""
    ks = 0
    ke = len(dec_list)
    if starttime is not None:
        # (because of tuple comparison, all bisects are left)
        # dec_list[:ks][0] < starttime <= dec_list[ks:][0]
        ks = bisect.bisect_left(dec_list, (starttime,))
        # forward fill, first index where dec_list[ks][0] <= starttime
        if ffill and (ks == len(dec_list) or dec_list[ks][0] > starttime):
            ks = max(ks - 1, 0)
    if endtime is not None:
        # (because of tuple comparison, all bisects are left)
        # dec_list[:ke][0] < endtime <= dec_list[ke:][0]
        ke = bisect.bisect_left(dec_list, (endtime,), lo=ks)
        # make ke first index to exclude
        if ke < len(dec_list) and dec_list[ke][0] == endtime:
            ke = ke + 1
    return slice(ks, ke)


def _yield_matching_files(
    root, dirs, props, include_drf, include_dmd, starttime=None, endtime=None,
    reverse=False,
):
    """Yield matching files from a list of subdirectories in a channel dir."""
    is_drf_channel = (
        any(_RE_DRFPROPFILE.match(f) for f in props) and include_drf
    )
    is_dmd_channel = (
        any(_RE_DMDPROPFILE.match(f) for f in props) and include_dmd
    )
    if is_drf_channel and is_dmd_channel:
        file_regex = _RE_FILE
    elif is_drf_channel:
        file_regex = _RE_DRFFILE
    elif include_dmd:
        file_regex = _RE_DMDFILE
    # get time-stamped subdirectories from dirs list
    dec_subdirs = []
    others = []
    for d in dirs:
        m = _RE_SUBDIR.match(d)
        if m:
            dt = datetime.datetime(
                year=int(m.group('year')),
                month=int(m.group('month')),
                day=int(m.group('day')),
                hour=int(m.group('hour')),
                minute=int(m.group('minute')),
                second=int(m.group('second')),
                tzinfo=pytz.utc,
            )
            time = dt - util.epoch
            dec_subdirs.append((time, d))
        else:
            others.append(d)
    # limit list of dirs for recursion by modifying in place
    dirs[:] = others

    # filter subdirectories by time
    dec_subdirs.sort()
    subdir_slice = _decorated_list_slice(
        dec_subdirs, starttime=starttime, endtime=endtime,
        ffill=True,
    )

    for k, (time, subdir) in (
        enumerate(dec_subdirs[subdir_slice])
        if not reverse else
        enumerate(list(reversed(dec_subdirs[subdir_slice])))
    ):
        # list potential files and get groups of all matching files
        subdir_files = os.listdir(os.path.join(root, subdir))
        dec_files = _decorate_drf_files(
            os.path.join(root, subdir), subdir_files, file_regex,
        )
        dec_files.sort()
        if ((k == 0) and is_dmd_channel and subdir_slice.start > 0 and
                starttime and dec_files[0][0] > starttime):
            # need to include files from earlier subdir so we can
            # forward fill and include the metadata file prior to starttime
            for k_subdir in range(subdir_slice.start - 1, -1, -1):
                prior_subdir = dec_subdirs[k_subdir][-1]
                prior_files = os.listdir(os.path.join(root, prior_subdir))
                dec_prior_files = _decorate_drf_files(
                    os.path.join(root, prior_subdir), prior_files, file_regex,
                )
                if dec_prior_files:
                    dec_prior_files.extend(dec_files)
                    dec_files = dec_prior_files
                    dec_files.sort()
                    break

        # filter by time and yield files
        # forward fill if metadata since we assume metadata for an older
        # time is valid until the next metadata sample
        slc = _decorated_list_slice(
            dec_files, starttime=starttime, endtime=endtime,
            ffill=(k == 0) and is_dmd_channel,
        )
        for dec_file in (
            dec_files[slc]
            if not reverse else
            reversed(dec_files[slc])
        ):
            yield dec_file[-1]


def ilsdrf(
    path, recursive=True, reverse=False, starttime=None, endtime=None,
    include_drf=True, include_dmd=True,
    include_drf_properties=None, include_dmd_properties=None,
):
    """Generator of Digital RF/Metadata files contained in a channel directory.

    Sub-directories will be traversed in alphabetical order and files from each
    directory will be yielded in Digital RF (time) order.


    Parameters
    ----------

    path : string
        Parent directory to list.

    recursive : bool
        If True, also list files recursively in channel subdirectories.

    reverse : bool
        If True, traverse directories and list files in those directories in
        reversed order.

    starttime : datetime.datetime
        Data covering this time or after will be included. This has no effect
        on property files.

    endtime : datetime.datetime
        Data covering this time or earlier will be included. This has no effect
        on property files.

    include_drf : bool
        If True, include Digital RF files in listing.

    include_dmd : bool
        If True, include Digital Metadata files in listing.

    include_drf_properties : bool | None
        If True, include the Digital RF properties file in listing.
        If None, use `include_drf` value.

    include_dmd_properties : bool | None
        If True, include the Digital Metadata properties file in listing.
        If None, use `include_dmd` value.


    Yields
    ------

    Digital RF/Metadata files contained in `path`.

    """
    # convert starttime and endtime to timedeltas for comparison
    if starttime is not None:
        if starttime.tzinfo is None:
            starttime = pytz.utc.localize(starttime)
        starttime = starttime - util.epoch
    if endtime is not None:
        if endtime.tzinfo is None:
            endtime = pytz.utc.localize(endtime)
        endtime = endtime - util.epoch

    if include_drf_properties is None:
        include_drf_properties = include_drf
    if include_dmd_properties is None:
        include_dmd_properties = include_dmd

    include_properties = True
    if include_drf_properties and include_dmd_properties:
        prop_regex = _RE_PROPFILE
    elif include_drf_properties:
        prop_regex = _RE_DRFPROPFILE
    elif include_dmd_properties:
        prop_regex = _RE_DMDPROPFILE
    else:
        include_properties = False

    path = os.path.abspath(path)
    if include_drf or include_dmd:
        # check if path is already a time-stamped subdir with files,
        # and yield the files if so
        root, subdir = os.path.split(path)
        if _RE_SUBDIR.match(subdir):
            any_props = [f for f in os.listdir(root) if _RE_PROPFILE.match(f)]
            if any_props:
                for f in _yield_matching_files(
                    root, [subdir], any_props, include_drf, include_dmd,
                    starttime=starttime, endtime=endtime, reverse=reverse,
                ):
                    yield f

    # now search for channel directories, starting with path
    for root, dirs, files in os.walk(path):
        # determine if root is a channel directory (i.e. has property file)
        any_props = [f for f in files if _RE_PROPFILE.match(f)]
        if any_props:
            if include_properties:
                # first yield property files
                props = [
                    os.path.join(root, f) for f in any_props
                    if prop_regex.match(f)
                ]
                props.sort(reverse=reverse)
                for prop in props:
                    yield prop

            if include_drf or include_dmd:
                # list, match, filter, and yield files from subdirectories
                for f in _yield_matching_files(
                    root, dirs, any_props, include_drf, include_dmd,
                    starttime=starttime, endtime=endtime, reverse=reverse,
                ):
                    yield f

        if recursive:
            # walk through (non-ts-subdir) directories in sorted order
            dirs.sort(reverse=reverse)
        else:
            # don't go further if we're not recursive
            del dirs[:]


def lsdrf(*args, **kwargs):
    """Get list of Digital RF/Metadata files contained in a channel directory.

    Sub-directories will be traversed in alphabetical order and files from each
    directory will be yielded in Digital RF (time) order.

    Parameters
    ----------

    path : string
        Parent directory to list.

    recursive : bool
        If True, also list files recursively in channel subdirectories.

    reverse : bool
        If True, traverse directories and list files in those directories in
        reversed order.

    starttime : datetime.datetime
        Data covering this time or after will be included. This has no effect
        on property files.

    endtime : datetime.datetime
        Data covering this time or earlier will be included. This has no effect
        on property files.

    include_drf : bool
        If True, include Digital RF files in listing.

    include_dmd : bool
        If True, include Digital Metadata files in listing.

    include_drf_properties : bool | None
        If True, include the Digital RF properties file in listing.
        If None, use `include_drf` value.

    include_dmd_properties : bool | None
        If True, include the Digital Metadata properties file in listing.
        If None, use `include_dmd` value.


    Returns
    -------

    List of Digital RF/Metadata files contained in `path`.

    """
    return list(ilsdrf(*args, **kwargs))


def _add_time_group(parser):
    timegroup = parser.add_argument_group(title='time')
    timegroup.add_argument(
        '-s', '--starttime', dest='starttime', default=None,
        help='''Data covering this time or after will be included (no effect
                on property files). The time can be given as a datetime string
                (e.g. in ISO8601 format: 2016-01-01T15:24:00Z) or a float
                giving a Unix timestamp in seconds. (default: %(default)s)''',
    )
    timegroup.add_argument(
        '-e', '--endtime', dest='endtime', default=None,
        help='''Data covering this time or before will be included (no effect
                on property files). The time can be given as a datetime string
                (e.g. in ISO8601 format: 2016-01-01T15:24:00Z), a float
                giving a Unix timestamp in seconds, or a '+' followed by a time
                in seconds that will be taken relative to `starttime`.
                (default: %(default)s)''',
    )
    return parser


def _add_include_group(parser):
    includegroup = parser.add_argument_group(title='include')
    drfgroup = includegroup.add_mutually_exclusive_group()
    dmdgroup = includegroup.add_mutually_exclusive_group()
    drfpropsgroup = includegroup.add_mutually_exclusive_group()
    dmdpropsgroup = includegroup.add_mutually_exclusive_group()

    drfgroup.add_argument(
        '--drf', dest='include_drf', action='store_true', default=True,
        help='''Include Digital RF files.
                (default: True)''',
    )
    drfpropsgroup.add_argument(
        '--drfprops', dest='include_drf_properties', action='store_true',
        default=None,
        help='''Include drf_properties.h5 files. If unset, use value of --drf.
                (default: None)''',
    )
    drfgroup.add_argument(
        '--nodrf', dest='include_drf', action='store_false',
        help='''Do not include Digital RF files.
                (default: False)''',
    )
    drfpropsgroup.add_argument(
        '--nodrfprops', dest='include_drf_properties', action='store_false',
        help='''Do not include drf_properties.h5 files.
                (default: False)''',
    )
    dmdgroup.add_argument(
        '--dmd', dest='include_dmd', action='store_true', default=True,
        help='''Include Digital Metadata files.
                (default: True)''',
    )
    dmdpropsgroup.add_argument(
        '--dmdprops', dest='include_dmd_properties', action='store_true',
        default=None,
        help='''Include dmd_properties.h5 files. If unset, use value of --dmd.
                (default: None)''',
    )
    dmdgroup.add_argument(
        '--nodmd', dest='include_dmd', action='store_false',
        help='''Do not include Digital Metadata files.
                (default: False)''',
    )
    dmdpropsgroup.add_argument(
        '--nodmdprops', dest='include_dmd_properties', action='store_false',
        help='''Do not include dmd_properties.h5 files.
                (default: False)''',
    )
    return parser


def _build_ls_parser(Parser, *args):
    desc = 'List Digital RF/Metadata files in a (channel) directory.'
    parser = Parser(*args, description=desc)

    parser.add_argument(
        'dirs', nargs='*', default='.',
        help='''Data directory to list. (default: %(default)s)''',
    )
    parser.add_argument(
        '-r', '--recursive', action='store_true',
        help='''Also list files recursively in channel subdirectories.
                (default: %(default)s)''',
    )
    parser.add_argument(
        '-R', '--reverse', action='store_true',
        help='''Traverse directories and list files in those directories in
                reversed order. (default: %(default)s)''',
    )
    parser.add_argument(
        '--abs', action='store_true',
        help='''Return absolute paths instead of relative paths.
                (default: %(default)s)''',
    )
    parser.add_argument(
        '--sortall', action='store_true',
        help='''Sort all files together before listing (normally only sorted
                within subdirectories). (default: %(default)s)''',
    )

    parser = _add_time_group(parser)
    parser = _add_include_group(parser)

    parser.set_defaults(func=_run_ls)

    return parser


def _run_ls(args):
    if args.starttime is not None:
        args.starttime = util.parse_identifier_to_time(args.starttime)
    if args.endtime is not None:
        args.endtime = util.parse_identifier_to_time(
            args.endtime, ref_datetime=args.starttime,
        )

    if args.abs:
        def fixpath(path, start=None):
            return path
    else:
        fixpath = os.path.relpath

    kwargs = vars(args).copy()
    del kwargs['func']
    del kwargs['dirs']
    del kwargs['abs']
    del kwargs['sortall']

    if args.sortall:
        files = []
        for d in args.dirs:
            files.extend(lsdrf(d, **kwargs))
        files.sort(key=sortkey_drf)
        print('\n'.join([fixpath(f, d) for f in files]))
    else:
        for d in args.dirs:
            for f in ilsdrf(d, **kwargs):
                print(fixpath(f, d))


def _add_srcdest_arguments(parser):
    parser.add_argument('src', help='Source directory.')
    parser.add_argument('dest', help='Destination directory.')

    parser.add_argument(
        '-c', '--channel', dest='chs', action='append', default=[],
        help='''Channel names to include from source directory.
                (default: Treat `src` as a channel directory)''',
    )
    parser.add_argument(
        '--only', dest='recursive', action='store_false', default=True,
        help='''Only copy the specified channel directory (do not recurse from
                `src` to find channel subdirectories).
                (default: False)''',
    )
    parser.add_argument(
        '-R', '--reverse', action='store_true',
        help='''Traverse directories and include files in those directories in
                reversed order. (default: %(default)s)''',
    )

    parser = _add_time_group(parser)
    parser = _add_include_group(parser)
    return parser


def _parse_srcdest_args(args):
    args.src = os.path.abspath(args.src)
    args.dest = os.path.abspath(args.dest)

    args.chs = [b.strip() for a in args.chs for b in a.strip().split(',')]
    args.srcdests = [
        (os.path.join(args.src, ch), os.path.join(args.dest, ch))
        for ch in args.chs
    ]
    if not args.srcdests:
        args.srcdests = [(args.src, args.dest)]

    if args.starttime is not None:
        args.starttime = util.parse_identifier_to_time(args.starttime)
    if args.endtime is not None:
        args.endtime = util.parse_identifier_to_time(
            args.endtime, ref_datetime=args.starttime,
        )

    kwargs = vars(args).copy()
    del kwargs['func']
    del kwargs['src']
    del kwargs['dest']
    del kwargs['chs']
    del kwargs['srcdests']

    return args, kwargs


def _build_cp_parser(Parser, *args):
    desc = 'Copy Digital RF/Metadata files from source to destination.'
    parser = Parser(*args, description=desc)
    parser = _add_srcdest_arguments(parser)
    parser.set_defaults(func=_run_cp)
    return parser


def _run_cp(args):
    args, kwargs = _parse_srcdest_args(args)
    for (src, dest) in args.srcdests:
        for srcpath in ilsdrf(src, **kwargs):
            destpath = os.path.join(dest, os.path.relpath(srcpath, src))
            destdir = os.path.dirname(destpath)
            if not os.path.exists(destdir):
                os.makedirs(destdir)
            shutil.copy2(srcpath, destpath)


def _build_mv_parser(Parser, *args):
    desc = 'Move Digital RF/Metadata files from source to destination.'
    parser = Parser(*args, description=desc)
    parser = _add_srcdest_arguments(parser)
    parser.set_defaults(func=_run_mv)
    return parser


def _run_mv(args):
    args, kwargs = _parse_srcdest_args(args)
    for (src, dest) in args.srcdests:
        for srcpath in ilsdrf(src, **kwargs):
            destpath = os.path.join(dest, os.path.relpath(srcpath, src))
            destdir = os.path.dirname(destpath)
            if not os.path.exists(destdir):
                os.makedirs(destdir)
            shutil.move(srcpath, destpath)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = _build_ls_parser(ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Utility functions for Digital RF and Digital Metadata."""
import ast
import datetime

import dateutil.parser
import numpy as np
import pytz
import six

__all__ = (
    'epoch', 'parse_identifier_to_sample', 'parse_identifier_to_time',
    'sample_to_datetime', 'samples_to_timedelta',
)


epoch = datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)


def parse_identifier_to_sample(iden, samples_per_second=None, ref_index=None):
    """Get a sample index from different forms of identifiers.

    Parameters
    ----------

    iden : None | int | float | string
        If None or '', None is returned to indicate that the index should
        be automatically determined.
        If an integer, it is returned as the sample index.
        If a float, it is interpreted as a UTC timestamp (seconds since epoch)
        and the corresponding sample index is returned.
        If a string, three forms are permitted:
            1) a string which can be evaluated to an integer/float and
                interpreted as above,
            2) a string beginning with '+' and followed by an integer
                (float) expression, interpreted as samples (seconds) from
                `ref_index`, and
            3) a time in ISO8601 format, e.g. '2016-01-01T16:24:00Z'

    samples_per_second : numpy.longdouble, required for float and time `iden`
        Sample rate in Hz used to convert a time to a sample index.

    ref_index : int/long, required for '+' string form of `iden`
        Reference index from which string `iden` beginning with '+' are
        offset.


    Returns
    -------

    sample_index : int | None
        Index to the identified sample given in the number of samples since
        the epoch (time_since_epoch*sample_per_second).

    """
    is_relative = False
    if iden is None or iden == '':
        return None
    elif isinstance(iden, six.string_types):
        if iden.startswith('+'):
            is_relative = True
            iden = iden.lstrip('+')
        try:
            # int or float
            iden = ast.literal_eval(iden)
        except (ValueError, SyntaxError):
            if is_relative:
                raise ValueError(
                    '"+" identifier must be followed by an integer or float.'
                )
            # convert datetime to float
            dt = dateutil.parser.parse(iden)
            if dt.tzinfo is None:
                # assume UTC if timezone was not specified in the string
                dt = pytz.utc.localize(dt)
            iden = (dt - epoch).total_seconds()

    if isinstance(iden, float):
        if samples_per_second is None:
            raise ValueError(
                'samples_per_second required when time identifier is used.'
            )
        idx = int(np.uint64(iden*samples_per_second))
    else:
        idx = int(iden)

    if is_relative:
        if ref_index is None:
            raise ValueError(
                'ref_index required when relative "+" identifier is used.'
            )
        return idx + ref_index
    else:
        return idx


def parse_identifier_to_time(iden, samples_per_second=None, ref_datetime=None):
    """Get a time from different forms of identifiers.

    Parameters
    ----------

    iden : None | float | string | int
        If None or '', None is returned to indicate that the time should
        be automatically determined.
        If a float, it is interpreted as a UTC timestamp (seconds since epoch)
        and the corresponding datetime is returned.
        If an integer, it is interpreted as a sample index when
        `samples_per_second` is not None and a UTC timestamp otherwise.
        If a string, three forms are permitted:
            1) a string which can be evaluated to an integer/float and
                interpreted as above,
            2) a string beginning with '+' and followed by an integer
                (float) expression, interpreted as samples (seconds) from
                `ref_time`, and
            3) a time in ISO8601 format, e.g. '2016-01-01T16:24:00Z'

    samples_per_second : numpy.longdouble, required for integer `iden`
        Sample rate in Hz used to convert a sample index to a time.

    ref_datetime : datetime, required for '+' string form of `iden`
        Reference time from which string `iden` beginning with '+' are
        offset. Must be timezone-aware.


    Returns
    -------

    dt : datetime
        Datetime object giving the indicated time.

    """
    is_relative = False
    if iden is None or iden == '':
        return None
    elif isinstance(iden, six.string_types):
        if iden.startswith('+'):
            is_relative = True
            iden = iden.lstrip('+')
        try:
            # int or float
            iden = ast.literal_eval(iden)
        except (ValueError, SyntaxError):
            if is_relative:
                raise ValueError(
                    '"+" identifier must be followed by an integer or float.'
                )
            # parse string to datetime
            dt = dateutil.parser.parse(iden)
            if dt.tzinfo is None:
                # assume UTC if timezone was not specified in the string
                dt = pytz.utc.localize(dt)
            return dt

    if isinstance(iden, float) or samples_per_second is None:
        td = datetime.timedelta(seconds=iden)
    else:
        td = samples_to_timedelta(iden, samples_per_second)

    if is_relative:
        if ref_datetime is None:
            raise ValueError(
                'ref_datetime required when relative "+" identifier is used.'
            )
        elif (not isinstance(ref_datetime, datetime.datetime)
                or ref_datetime.tzinfo is None):
            raise ValueError(
                'ref_datetime must be a timezone-aware datetime.'
            )
        return td + ref_datetime
    else:
        return td + epoch


def sample_to_datetime(sample, samples_per_second):
    """Get datetime corresponding to the given sample index.

    Parameters
    ----------

    sample : int
        Sample index in number of samples since epoch.

    samples_per_second : numpy.longdouble
        Sample rate in Hz.


    Returns
    -------

    dt : datetime
        Datetime corresponding to the given sample index.

    """
    return epoch + samples_to_timedelta(sample, samples_per_second)


def samples_to_timedelta(samples, samples_per_second):
    """Get timedelta for a duration in number of samples given a sample rate.

    Parameters
    ----------

    samples : int
        Duration in number of samples.

    samples_per_second : numpy.longdouble
        Sample rate in Hz.


    Returns
    -------

    td : datetime.timedelta
        Timedelta corresponding to the number of samples.

    """
    # splitting into secs/frac lets us get a more accurate datetime
    secs = int(samples // samples_per_second)
    frac = (samples % samples_per_second) / samples_per_second
    microseconds = int(np.uint64(frac*1000000))

    return datetime.timedelta(seconds=secs, microseconds=microseconds)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Module for mirroring Digital RF files from on directory to another."""

import filecmp
import os
import shutil
import sys
import time
import traceback
from itertools import chain

from watchdog.events import FileCreatedEvent

from . import list_drf, util
from .list_drf import ilsdrf
from .ringbuffer import DigitalRFRingbufferHandler
from .watchdog_drf import DigitalRFEventHandler, DirWatcher

__all__ = (
    'DigitalRFMirrorHandler', 'DigitalRFMirror',
)


class DigitalRFMirrorHandler(DigitalRFEventHandler):
    """Event handler for mirroring Digital RF and Digital Metadata files.

    This handler mirrors new or modified files from a source directory to a
    destination directory. Moved and deleted files are ignored.

    """

    def __init__(
        self, src, dest, verbose=False, mirror_fun=shutil.copy2,
        starttime=None, endtime=None, include_drf=True, include_dmd=True,
        include_drf_properties=True, include_dmd_properties=True,
    ):
        """Create Digital RF mirror handler given source and destination.

        Other Parameters
        ----------------

        starttime : datetime.datetime
            Data covering this time or after will be included. This has no
            effect on property files.

        endtime : datetime.datetime
            Data covering this time or earlier will be included. This has no
            effect on property files.

        include_drf : bool
            If True, include Digital RF files.

        include_dmd : bool
            If True, include Digital Metadata files.

        include_drf_properties : bool | None
            If True, include the Digital RF properties file.
            If None, use `include_drf` value.

        include_dmd_properties : bool | None
            If True, include the Digital Metadata properties file.
            If None, use `include_dmd` value.

        """
        self.src = os.path.abspath(src)
        self.dest = os.path.abspath(dest)
        self.verbose = verbose
        self.mirror_fun = mirror_fun
        super(DigitalRFMirrorHandler, self).__init__(
            starttime=starttime, endtime=endtime,
            include_drf=include_drf, include_dmd=include_dmd,
            include_drf_properties=include_drf_properties,
            include_dmd_properties=include_dmd_properties,
        )

    def _get_dest_path(self, src_path):
        rel_path = os.path.relpath(src_path, self.src)
        dest_path = os.path.join(self.dest, rel_path)
        return dest_path

    def mirror_to_dest(self, src_path):
        """Mirror file to its location in the destination directory."""
        dest_path = self._get_dest_path(src_path)
        dest_dir, dest_name = os.path.split(dest_path)
        tmp_dest_path = os.path.join(dest_dir, 'tmp.' + dest_name)
        try:
            if not os.path.exists(dest_dir):
                os.makedirs(dest_dir)
            if (not os.path.exists(dest_path)
                    or not filecmp.cmp(src_path, dest_path)):
                if self.verbose:
                    print('Mirroring {0}'.format(src_path))
                else:
                    sys.stdout.write('.')
                    sys.stdout.flush()
                # mirror to temporary name, then rename to final destination
                self.mirror_fun(src_path, tmp_dest_path)
                shutil.move(tmp_dest_path, dest_path)
        except OSError:
            if not os.path.isfile(src_path):
                # file doesn't exist anymore, no need to notify
                pass
            else:
                # otherwise, print the error but don't stop mirroring
                traceback.print_exc()

        # try to clean up source directory in case it is empty
        src_dir, src_name = os.path.split(src_path)
        try:
            os.rmdir(src_dir)
        except OSError:
            # directory not empty, just move on
            pass

    def on_created(self, event):
        """Mirror newly-created file."""
        self.mirror_to_dest(event.src_path)

    def on_modified(self, event):
        """Mirror modified file."""
        self.mirror_to_dest(event.src_path)


class DigitalRFMirror(object):
    """Monitor a directory and mirror its Digital RF files to another.

    This class combines an event handler and a file system observer. It
    monitors a source directory and mirrors new and modified (and optionally,
    existing) Digital RF and Digital Metadata files to a given destination
    directory. Files that are moved or deleted in the source directory are
    ignored.

    """

    def __init__(
        self, src, dest, method='move', ignore_existing=False, verbose=False,
        starttime=None, endtime=None, include_drf=True, include_dmd=True,
    ):
        """Create Digital RF mirror object. Use start/run method to begin.

        Parameters
        ----------

        src : str
            Source directory to monitor.

        dest : str
            Destination directory.

        method : 'move' | 'copy'
            Mirroring method. New Digital RF files in the source directory will
            be moved or copied to the destination directory depending on this
            parameter. Metadata is always copied regardless of this parameter.


        Other Parameters
        ----------------

        ignore_existing : bool
            If True, existing Digital RF and Digital Metadata files will not be
            mirrored. Otherwise, they will be mirrored to the destination
            directory along with any existing channel metadata files
            when mirroring starts.

        verbose : bool
            If True, print the name of mirrored files.

        starttime : datetime.datetime
            Data covering this time or after will be included. This has no
            effect on property files.

        endtime : datetime.datetime
            Data covering this time or earlier will be included. This has no
            effect on property files.

        include_drf : bool
            If True, include Digital RF files. If False, ignore Digital RF
            files.

        include_dmd : bool
            If True, include Digital Metadata files. If False, ignore Digital
            Metadata files.

        """
        self.src = os.path.abspath(src)
        self.dest = os.path.abspath(dest)
        if method not in ('move', 'copy'):
            raise ValueError('Mirror method must be either "move" or "copy".')
        self.method = method
        self.ignore_existing = ignore_existing
        self.verbose = verbose
        self.starttime = starttime
        self.endtime = endtime
        self.include_drf = include_drf
        self.include_dmd = include_dmd

        if not self.include_drf and not self.include_dmd:
            errstr = 'One of `include_drf` or `include_dmd` must be True.'
            raise ValueError(errstr)

        self.event_handlers = []
        # have to copy properties files because static,
        # have to copy metadata because can be modified
        copy_handler = DigitalRFMirrorHandler(
            self.src, self.dest, verbose=verbose, mirror_fun=shutil.copy2,
            starttime=self.starttime, endtime=self.endtime,
            include_drf=(self.include_drf and self.method == 'copy'),
            include_dmd=self.include_dmd,
            include_drf_properties=self.include_drf,
            include_dmd_properties=self.include_dmd,
        )
        self.event_handlers.append(copy_handler)

        if self.include_drf and self.method == 'move':
            # move RF files with a separate handler
            drf_handler = DigitalRFMirrorHandler(
                self.src, self.dest, verbose=verbose, mirror_fun=shutil.move,
                starttime=self.starttime, endtime=self.endtime,
                include_drf=True, include_dmd=False,
                include_drf_properties=False, include_dmd_properties=False,
            )
            self.event_handlers.append(drf_handler)

        if self.include_dmd and self.method == 'move':
            # set ringbuffer on Digital Metadata files so old ones are removed
            # (can't move since multiple writes can happen to a single file)
            md_ringbuffer_handler = DigitalRFRingbufferHandler(
                count=1, verbose=verbose, dryrun=False,
                starttime=self.starttime, endtime=self.endtime,
                include_drf=False, include_dmd=True,
            )
            self.event_handlers.append(md_ringbuffer_handler)

        self._init_observer()

    def _init_observer(self):
        self.observer = DirWatcher(self.src)
        for handler in self.event_handlers:
            self.observer.schedule(handler, self.src, recursive=True)

    def start(self):
        """Start mirror process and return when existing files are handled."""
        # start observer to mirror new and modified files
        self.observer.start()

        if os.path.isdir(self.src):
            # don't need to pause dispatching because mirror ordering is not
            # critical and duplicate events are not harmful (we will either
            # copy again or fail to move because the source doesn't exist)
            # mirror properties at minimum
            paths = ilsdrf(
                self.src, include_drf=False, include_dmd=False,
                include_drf_properties=self.include_drf,
                include_dmd_properties=self.include_dmd,
            )

            if not self.ignore_existing:
                # mirror other files if desired
                more_paths = ilsdrf(
                    self.src, starttime=self.starttime, endtime=self.endtime,
                    include_drf=self.include_drf,
                    include_dmd=self.include_dmd,
                    include_drf_properties=False,
                    include_dmd_properties=False,
                )
                paths = chain(paths, more_paths)

            # add events for existing files, skipping observer and event queue
            # and dispatching directly so we can dispatch from this thread
            # and ignore file time (which ilsdrf has already checked, correctly
            # matching most recent DMD file at or before starttime)
            for p in paths:
                event = FileCreatedEvent(p)
                for handler in self.event_handlers:
                    handler.dispatch(event, match_time=False)

    def join(self):
        """Wait until a KeyboardInterrupt is received to stop mirroring."""
        try:
            while True:
                if not self.observer.all_alive():
                    # if not all threads of the observer are alive,
                    # reinitialize and restart
                    print(
                        'Found stopped thread, reinitializing and restarting.'
                    )
                    # make a new observer and start it ASAP
                    # (if we missed some events, can't help it)
                    self._init_observer()
                    self.observer.start()
                time.sleep(1)
        except (KeyboardInterrupt, SystemExit):
            self.stop()
            sys.stdout.write('\n')
            sys.stdout.flush()
        self.observer.join()

    def run(self):
        """Start mirroring and wait for a KeyboardInterrupt to stop."""
        self.start()
        self.join()

    def stop(self):
        """Stop mirror process."""
        self.observer.stop()


def _build_mirror_parser(Parser, *args):
    desc = 'Mirror Digital RF files from one directory to another.'
    parser = Parser(*args, description=desc)

    parser.add_argument(
        'method', choices=['mv', 'cp'],
        help='Mirroring method.',
    )
    parser.add_argument('src', help='Source directory to monitor.')
    parser.add_argument('dest', help='Destination directory.')
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Print the name of mirrored files.',
    )
    parser.add_argument(
        '--ignore_existing', action='store_true',
        help='Ignore existing files in source directory.',
    )

    parser = list_drf._add_time_group(parser)

    includegroup = parser.add_argument_group(title='include')
    includegroup.add_argument(
        '--nodrf', dest='include_drf', action='store_false',
        help='''Do not mirror Digital RF HDF5 files.
                (default: False)''',
    )
    includegroup.add_argument(
        '--nodmd', dest='include_dmd', action='store_false',
        help='''Do not mirror Digital Metadata HDF5 files.
                (default: False)''',
    )

    parser.set_defaults(func=_run_mirror)

    return parser


def _run_mirror(args):
    methods = {'mv': 'move', 'cp': 'copy'}
    args.method = methods[args.method]

    if args.starttime is not None:
        args.starttime = util.parse_identifier_to_time(args.starttime)
    if args.endtime is not None:
        args.endtime = util.parse_identifier_to_time(
            args.endtime, ref_datetime=args.starttime,
        )

    kwargs = vars(args).copy()
    del kwargs['func']
    mirror = DigitalRFMirror(**kwargs)
    print('Mirroring ({0}) {1} to {2}.'.format(
        args.method, args.src, args.dest,
    ))
    print('Type Ctrl-C to quit.')
    mirror.run()


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = _build_mirror_parser(ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Module for montoring creation/modification/deletion of Digital RF files."""

import datetime
import os
import re
import sys
import time
from contextlib import contextmanager

import pytz
from watchdog.events import (DirCreatedEvent, FileCreatedEvent,
                             FileDeletedEvent, RegexMatchingEventHandler)
from watchdog.observers import Observer
from watchdog.observers.api import ObservedWatch
from watchdog.utils import unicode_paths
from watchdog.utils.bricks import OrderedSetQueue

from . import list_drf, util
from .list_drf import (RE_DMD, RE_DMDPROP, RE_DRF, RE_DRFDMD, RE_DRFDMDPROP,
                       RE_DRFPROP)

__all__ = (
    'DigitalRFEventHandler', 'DirWatcher',
)


class DigitalRFEventHandler(RegexMatchingEventHandler):
    """Event handler for Digital RF and Digital Metadata files.

    Events to files that match the Digital RF or Digital Metadata format and
    directory structure will trigger the corresponding method. Override these
    methods in a subclass to define their actions.

    The triggered methods are: `on_created`, `on_deleted`, `on_modified`, and
    `on_moved`.

    """

    def __init__(
        self, starttime=None, endtime=None, include_drf=True, include_dmd=True,
        include_drf_properties=None, include_dmd_properties=None,
        ignore_regexes=[],
    ):
        """Create Digital RF event handler for given time range and file types.

        Parameters
        ----------

        starttime : datetime.datetime
            Data covering this time or after will be included. This has no
            effect on property files.

        endtime : datetime.datetime
            Data covering this time or earlier will be included. This has no
            effect on property files.

        include_drf : bool
            If True, include Digital RF files.

        include_dmd : bool
            If True, include Digital Metadata files.

        include_drf_properties : bool | None
            If True, include the Digital RF properties file.
            If None, use `include_drf` value.

        include_dmd_properties : bool | None
            If True, include the Digital Metadata properties file.
            If None, use `include_dmd` value.

        ignore_regexes : list
            List of regexes for ignoring matching event paths.

        """
        # convert starttime and endtime to timedeltas for comparison
        if starttime is not None:
            if starttime.tzinfo is None:
                starttime = pytz.utc.localize(starttime)
            starttime = starttime - util.epoch
        self.starttime = starttime
        if endtime is not None:
            if endtime.tzinfo is None:
                endtime = pytz.utc.localize(endtime)
            endtime = endtime - util.epoch
        self.endtime = endtime

        if include_drf_properties is None:
            include_drf_properties = include_drf
        if include_dmd_properties is None:
            include_dmd_properties = include_dmd

        regexes = []
        if include_drf and include_dmd:
            regexes.append(RE_DRFDMD)
        elif include_drf:
            regexes.append(RE_DRF)
        elif include_dmd:
            regexes.append(RE_DMD)
        if include_drf_properties and include_dmd_properties:
            regexes.append(RE_DRFDMDPROP)
        elif include_drf_properties:
            regexes.append(RE_DRFPROP)
        elif include_dmd_properties:
            regexes.append(RE_DMDPROP)

        if not regexes:
            raise ValueError('Must include at least one file type.')

        super(DigitalRFEventHandler, self).__init__(
            regexes=regexes, ignore_regexes=ignore_regexes,
            ignore_directories=True,
        )

    def dispatch(self, event, match_time=True):
        """Dispatch events to the appropriate methods.

        Parameters
        ----------

        event : FileSystemEvent
            Event object representing the file system event.

        match_time : bool
            If False, do not check the matched file's time against the
            handler's starttime and endtime.

        """
        if self.ignore_directories and event.is_directory:
            return

        src_match = False
        if event.src_path:
            src_path = unicode_paths.decode(event.src_path)
            if not any(r.match(src_path) for r in self.ignore_regexes):
                for r in self.regexes:
                    m = r.match(src_path)
                    if m:
                        src_match = True
                        match = m

        dest_match = False
        if getattr(event, 'dest_path', None) is not None:
            dest_path = unicode_paths.decode(event.dest_path)
            if not any(r.match(dest_path) for r in self.ignore_regexes):
                for r in self.regexes:
                    m = r.match(dest_path)
                    if m:
                        dest_match = True
                        match = m

            # change move event to deleted/created if both regexes didn't match
            if src_match and not dest_match:
                event = FileDeletedEvent(event.src_path)
            elif dest_match and not src_match:
                event = FileCreatedEvent(event.dest_path)

        if not src_match and not dest_match:
            return

        # regexes matched, now check the time
        if match_time:
            try:
                secs = int(match.group('secs'))
            except (IndexError, TypeError):
                # no time, don't need to check it
                pass
            else:
                try:
                    msecs = int(match.group('frac'))
                except (IndexError, TypeError):
                    msecs = 0
                time = datetime.timedelta(seconds=secs, milliseconds=msecs)
                if self.starttime is not None and time < self.starttime:
                    return
                elif self.endtime is not None and time > self.endtime:
                    return

        # the event matched, including time if applicable, dispatch
        self.on_any_event(event)
        _method_map = {
            'modified': self.on_modified,
            'moved': self.on_moved,
            'created': self.on_created,
            'deleted': self.on_deleted,
        }
        event_type = event.event_type
        _method_map[event_type](event)


class DirWatcher(Observer, RegexMatchingEventHandler):
    """Watchdog observer for monitoring a particular directory.

    This observer has a sub-observer and handler for noticing when the
    specified directory is created or deleted so that observations of that
    directory can be preserved. In this way observations can be scheduled
    for a directory that doesn't yet exist, and they will be enacted once it is
    created.

    """

    def __init__(self, path, **kwargs):
        """Create observer for the directory at `path`."""
        Observer.__init__(self, **kwargs)
        # replace default event queue with ordered set queue to disallow
        # duplicate events even if added out of order
        self._event_queue = OrderedSetQueue()
        RegexMatchingEventHandler.__init__(self)

        self.path = os.path.abspath(path)
        self.root_observer = Observer(**kwargs)
        self.root_watch = None
        self._stopped_handlers = dict()
        self._dispatching_enabled = True

    def _get_next_dir_in_path(self, start):
        """From `start`, find the next directory from self.path."""
        relpath = os.path.relpath(self.path, start)
        nextdirname = relpath.split(os.path.sep)[0]
        nextdir = os.path.abspath(os.path.join(start, nextdirname))
        return nextdir

    def _set_root(self, root):
        """Set up watching `root` or closest existing parent."""
        # schedule new root watch
        while True:
            try:
                watch = self.root_observer.schedule(
                    event_handler=self, path=root, recursive=False,
                )
            except OSError:
                # root doesn't exist, move up one directory and try again
                try:
                    # clean up from failed scheduling
                    self.root_observer.unschedule(
                        ObservedWatch(root, False),
                    )
                except KeyError:
                    pass
                root = os.path.dirname(root)
            else:
                # watch was set, break out of while loop
                break

        # add regex for root and next subdirectory
        regexes = []
        regexes.append('^' + re.escape(root) + '$')
        nextdir = self._get_next_dir_in_path(root)
        if nextdir != root:
            regexes.append('^' + re.escape(nextdir) + '$')

        # update handler (self) with regexes
        RegexMatchingEventHandler.__init__(
            self, regexes=regexes, ignore_directories=False,
        )

        # unschedule old root watch
        if self.root_watch is not None:
            try:
                self.root_observer.unschedule(self.root_watch)
            except KeyError:
                # emitter already stopped
                pass

        self.root_watch = watch

    def _start_dispatching(self):
        self._dispatching_enabled = True

    def _stop_dispatching(self):
        # wait until event queue is cleared
        self.event_queue.join()
        self._dispatching_enabled = False

    def _start_watching_path(self):
        """Schedule handlers for self.path now that it exists."""
        for watch, handlers in self._stopped_handlers.items():
            for handler in handlers:
                self.schedule(
                    event_handler=handler, path=watch.path,
                    recursive=watch.is_recursive
                )
        # generate any events for files/dirs in self.path that were
        # created before the watch started (since dispatching is stopped,
        # duplicate events will be caught and ignored)
        with self._lock:
            for emitter in self.emitters:
                dirs = [emitter.watch.path]
                for root in dirs:
                    names = os.listdir(root)
                    paths = [os.path.join(root, name) for name in names]
                    paths.sort(key=os.path.getmtime)
                    for path in paths:
                        if os.path.isdir(path):
                            if emitter.watch.is_recursive:
                                dirs.append(path)
                            event = DirCreatedEvent(path)
                        else:
                            event = FileCreatedEvent(path)
                        emitter.queue_event(event)
        # start dispatching events
        self._start_dispatching()

    def _stop_watching_path(self):
        """Save handlers for self.path so they can be reinstated, clean up."""
        # stop further event dispatching
        self._stop_dispatching()
        with self._lock:
            # copy all watches/handlers so we can restart them
            self._stopped_handlers.update(self._handlers.copy())
        for watch in list(self._watches):
            try:
                self.unschedule(watch)
            except KeyError:
                # emitter already stopped
                pass
        self._watches.clear()

    def on_created(self, event):
        """Adjust watches for just-created directory in self.path hierarchy."""
        if event.is_directory:
            # directory in self.path tree has been created
            # add watch for directory
            self._set_root(event.src_path)
            if event.src_path == self.path:
                # if it is self.path, start watching
                self._start_watching_path()
            else:
                # generate event for sub-directory in path that was
                # created before the watch started
                nextdir = self._get_next_dir_in_path(event.src_path)
                if os.path.isdir(nextdir):
                    emitter = self.root_observer._emitter_for_watch[
                        self.root_watch
                    ]
                    event = DirCreatedEvent(nextdir)
                    emitter.queue_event(event)

    def on_deleted(self, event):
        """Adjust watches for deleted directory in self.path hierarchy."""
        if event.is_directory:
            # root directory in self.path tree has been deleted
            # stop watching if it is self.path
            if event.src_path == self.path:
                self._stop_watching_path()
            # set root to (closest) parent of deleted directory
            self._set_root(os.path.dirname(event.src_path))

    def schedule(self, event_handler, path=None, recursive=False):
        """Schedule watching a path with an event handler."""
        if path is None:
            path = self.path
        try:
            super(DirWatcher, self).schedule(
                event_handler=event_handler, path=path, recursive=recursive,
            )
        except OSError:
            # path doesn't exist and we're already running, but watch was added
            # to _handlers, so re-do _stop_watching_path
            self._stop_watching_path()

    def start(self):
        """Start watching and enable handlers."""
        # start observing directories
        self.root_observer.start()
        # start observer before adding path so it can find closest root
        self._set_root(self.path)
        # start self observer
        if not os.path.isdir(self.path):
            # save scheduled watches so we can start without error
            self._stop_watching_path()
        super(DirWatcher, self).start()

    def all_alive(self):
        """Check if all observer and emitter threads are running."""
        # check if self thread is running
        if not self.is_alive():
            return False
        # check if self emitters are running
        if not all(
            emitter.is_alive()
            and (not emitter._inotify or emitter._inotify.is_alive())
                for emitter in self.emitters
        ):
            return False
        # check if root observer thread is running
        if not self.root_observer.is_alive():
            return False
        # check if all root observer emitters are running
        if not all(
            emitter.is_alive()
            and (not emitter._inotify or emitter._inotify.is_alive())
                for emitter in self.root_observer.emitters
        ):
            return False

        return True

    def dispatch_events(self, event_queue, timeout):
        """Get events from queue and dispatch them to handlers."""
        # override this so that we can schedule without dispatching events
        # immediately even while thread is running
        if self._dispatching_enabled:
            super(DirWatcher, self).dispatch_events(event_queue, timeout)
        else:
            time.sleep(timeout)

    @contextmanager
    def paused_dispatching(self):
        """Context manager that pauses event dispatching while held."""
        self._stop_dispatching()
        yield
        self._start_dispatching()


def _build_watch_parser(Parser, *args):
    desc = 'Print Digital RF file events occurring in a directory.'
    parser = Parser(*args, description=desc)

    parser.add_argument('dir', nargs='?', default='.',
                        help='''Data directory to monitor.
                               (default: %(default)s)''')

    parser = list_drf._add_time_group(parser)
    parser = list_drf._add_include_group(parser)

    parser.set_defaults(func=_run_watch)

    return parser


def _run_watch(args):
    args.dir = os.path.abspath(args.dir)

    if args.starttime is not None:
        args.starttime = util.parse_identifier_to_time(args.starttime)
    if args.endtime is not None:
        args.endtime = util.parse_identifier_to_time(
            args.endtime, ref_datetime=args.starttime,
        )

    # subclass DigitalRFEventHandler to just print events
    class DigitalRFPrint(DigitalRFEventHandler):

        def on_created(self, event):
            print('Created {0}'.format(event.src_path))

        def on_deleted(self, event):
            print('Deleted {0}'.format(event.src_path))

        def on_modified(self, event):
            print('Modified {0}'.format(event.src_path))

        def on_moved(self, event):
            print('Moved {0} to {1}'.format(event.src_path, event.dest_path))

    kwargs = vars(args).copy()
    del kwargs['func']
    del kwargs['dir']
    event_handler = DigitalRFPrint(**kwargs)
    observer = DirWatcher(args.dir)
    observer.schedule(event_handler, args.dir, recursive=True)
    print('Monitoring {0}:'.format(args.dir))
    observer.start()
    try:
        while True:
            time.sleep(1)
    except (KeyboardInterrupt, SystemExit):
        observer.stop()
        sys.stdout.write('\n')
        sys.stdout.flush()
    observer.join()


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = _build_watch_parser(ArgumentParser)
    args = parser.parse_args()
    args.func(args)



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""

drf_sti_summary.py
$Id$

Create a spectral time intensity summary plot for a data set.

"""


import traceback
import sys
import string
import optparse
import os

import matplotlib.pyplot
import matplotlib.mlab
import matplotlib.gridspec

import scipy
import scipy.signal

import numpy
import numpy.fft
import time
import datetime
import dateutil
import pytz

import digital_rf as drf
matplotlib.rc('axes', hold=False)


class DataPlotter:

    def __init__(self, control):
        """__init__: Initializes a data plotter for STI plotting.

        """
        self.control = control
        ch = string.split(self.control.channel, ':')
        self.channel = ch[0]
        self.sub_channel = int(ch[1])

        # open digital RF path
        self.dio = drf.DigitalRFReader(self.control.path)

        if self.control.verbose:
            print 'channel bounds:', self.dio.get_bounds(self.channel)

        self.bounds = self.dio.get_bounds(self.channel)

        print 'bounds ', self.bounds

        # Figure setup

        self.f = matplotlib.pyplot.figure(figsize=(7, numpy.min(
            [numpy.max([4, self.control.frames]), 7])), dpi=128)

        self.gridspec = matplotlib.gridspec.GridSpec(self.control.frames, 1)

        self.subplots = []

        """ Setup the subplots for this display """
        for n in numpy.arange(self.control.frames):
            ax = self.f.add_subplot(self.gridspec[n])
            self.subplots.append(ax)

    def plot(self):
        """

            Iterate over the data set and plot the STI into the subplot panels. Each
            panel is divided into a provided number of bins of a given integration
            length. Strides between the panels are made between integrations.

        """

        # initialize outside the loop to avoid memory leak

        plot_a = None

        # initial plotting scales
        vmin = 0
        vmax = 0
        pmin = 0
        pmax = 0

        sr = self.dio.get_properties(self.channel)['samples_per_second']

        if self.control.verbose:
            print 'sample rate: ', sr

        # initial time info
        display_lag = 60
        b = self.dio.get_bounds(self.channel)

        if self.control.verbose:
            print 'data bounds: ', b

        if self.control.start:
            dtst0 = dateutil.parser.parse(self.control.start)
            st0 = (dtst0 - datetime.datetime(1970, 1,
                                             1, tzinfo=pytz.utc)).total_seconds()
            st0 = int(st0 * sr)
        else:
            st0 = int(b[0])

        if self.control.end:
            dtst0 = dateutil.parser.parse(self.control.end)
            et0 = (dtst0 - datetime.datetime(1970, 1,
                                             1, tzinfo=pytz.utc)).total_seconds()
            et0 = int(et0 * sr)
        else:
            et0 = int(b[1])

        if self.control.verbose:

            print 'start sample st0: ', st0
            print 'end sample et0: ', et0

        blocks = self.control.bins * self.control.frames

        samples_per_stripe = self.control.num_fft * \
            self.control.integration * self.control.decimation
        total_samples = blocks * samples_per_stripe

        if total_samples > (et0 - st0):
            print 'Insufficient samples for %d samples per stripe and %d blocks between %ld and %ld' % (samples_per_stripe, blocks, st0, et0)
            return

        stripe_stride = (et0 - st0) / blocks

        bin_stride = stripe_stride / self.control.bins

        start_sample = st0

        print 'first ', start_sample

        # get metadata
        # this could be done better to ensure we catch frequency or sample rate
        # changes
        mdt = self.dio.read_metadata(st0, et0, self.channel)
        try:
            md = mdt[mdt.keys()[0]]
            cfreq = md['center_frequencies'].ravel()[self.sub_channel]
        except (IndexError, KeyError):
            cfreq = 0.0

        if self.control.verbose:
            print 'processing info : ', self.control.frames, self.control.bins, samples_per_stripe, bin_stride

        for p in numpy.arange(self.control.frames):
            sti_psd_data = numpy.zeros(
                [self.control.num_fft, self.control.bins], numpy.float)
            sti_times = numpy.zeros([self.control.bins], numpy.complex128)

            for b in numpy.arange(self.control.bins):

                if self.control.verbose:
                    print 'read vector :', self.channel, start_sample, samples_per_stripe

                d_vec = self.dio.read_vector(
                    start_sample, samples_per_stripe, self.channel)
                data = d_vec[:, self.sub_channel]

                if self.control.decimation > 1:
                    data = scipy.signal.decimate(data, self.control.decimation)
                    sample_freq = sr / self.control.decimation
                else:
                    sample_freq = sr

                if self.control.mean:
                    detrend_fn = matplotlib.mlab.detrend_mean
                else:
                    detrend_fn = matplotlib.mlab.detrend_none

                try:
                    psd_data, freq_axis = matplotlib.mlab.psd(
                        data, NFFT=self.control.num_fft, Fs=float(sample_freq), detrend=detrend_fn, scale_by_freq=False)
                except:
                    traceback.print_exc(file=sys.stdout)

                sti_psd_data[:, b] = numpy.real(
                    10.0 * numpy.log10(numpy.abs(psd_data) + 1E-12))

                sti_times[b] = start_sample / sr

                start_sample += stripe_stride

            # Now Plot the Data
            ax = self.subplots[p]

            # determine image x-y extent
            extent = (
                0,
                self.control.bins,
                numpy.min(freq_axis) / 1e3,
                numpy.max(freq_axis) / 1e3,
            )

            # determine image color extent in log scale units
            Pss = sti_psd_data
            vmin = numpy.real(numpy.median(Pss) - 6.0)
            vmax = numpy.real(numpy.median(
                Pss) + (numpy.max(Pss) - numpy.median(Pss)) * 0.61803398875 + 50.0)

            if self.control.zaxis:
                vmin = int(string.split(self.control.zaxis, ':')[0])
                vmax = int(string.split(self.control.zaxis, ':')[1])
            else:
                vmin = numpy.real(numpy.median(Pss) - 6.0)
                vmax = numpy.real(numpy.median(
                    Pss) + (numpy.max(Pss) - numpy.median(Pss)) * 0.61803398875 + 50.0)

            im = ax.imshow(sti_psd_data, cmap='jet', origin='lower', extent=extent,
                           interpolation='nearest', vmin=vmin, vmax=vmax, aspect='auto')

            ax.set_ylabel('f (kHz)', fontsize=8)

            # plot dates

            tick_spacing = numpy.arange(
                self.control.bins / 8, self.control.bins, self.control.bins / 8)
            ax.set_xticks(tick_spacing)
            tick_labels = []

            for s in tick_spacing:
                tick_time = sti_times[s]

                if tick_time == 0:
                    tick_string = ''
                else:
                    gm_tick_time = time.gmtime(numpy.real(tick_time))
                    tick_string = '%02d:%02d:%02d' % (
                        gm_tick_time[3], gm_tick_time[4], gm_tick_time[5])
                    tick_labels.append(tick_string)

            ax.set_xticklabels(tick_labels)

            # set the font sizes
            tl = ax.get_xticklabels()

            for tk in tl:
                tk.set_size(8)
            del tl

            tl = ax.get_yticklabels()

            for tk in tl:
                tk.set_size(8)
            del tl

        print 'last ', start_sample

        # create a time stamp
        start_time = st0 / sr
        srt_time = time.gmtime(start_time)
        sub_second = int(round((start_time - int(start_time)) * 100))

        timestamp = "%d-%02d-%02d %02d:%02d:%02d.%02d UT" % (srt_time[0], srt_time[
                                                             1], srt_time[2], srt_time[3], srt_time[4], srt_time[5], sub_second)

        self.f.suptitle('%s %s %4.2f MHz (%s)' % (
            self.control.title, timestamp, cfreq / 1E6, self.control.path), fontsize=10)

        # ax.legend(fontsize=8)
        ax.set_xlabel('time (UTC)', fontsize=8)

        # fixup ticks

        tl = ax.get_xticklabels()
        for tk in tl:
            tk.set_size(8)
        del tl
        tl = ax.get_yticklabels()
        for tk in tl:
            tk.set_size(8)
        del tl

        self.gridspec.update()

        self.f.tight_layout()

        self.f.subplots_adjust(top=0.95, right=0.88)
        cax = self.f.add_axes([0.9, 0.12, 0.02, 0.80])
        self.f.colorbar(im, cax=cax)
        if self.control.outname:
            fname, ext = os.path.splitext(self.control.outname)
            if ext == '':
                ext = '.png'
            print "Save plot as {}".format(fname+ext)
            matplotlib.pyplot.savefig(fname+ext)
        if self.control.appear or not self.control.outname:
            print "Show plot"
            matplotlib.pyplot.show()


def parse_command_line(str_input=None):
    # if str_input is None:
    #     parser = optparse.OptionParser()
    # else:
    #     parser = optparse.OptionParser(str_input)
    parser = optparse.OptionParser()

    parser.add_option("-t", "--title", dest="title",
                      default='Digital RF Data', help="Use title provided for the data.")
    parser.add_option("-s", "--start", dest="start", default=None,
                      help="Use the provided start time instead of the first time in the data. format is ISO8601: 2015-11-01T15:24:00Z")
    parser.add_option("-e", "--end", dest="end", default=None,
                      help="Use the provided end time for the plot. format is ISO8601: 2015-11-01T15:24:00Z")

    parser.add_option("-p", "--path", dest="path",
                      help="Use data from the provided digital RF data <path>.")
    parser.add_option("-c", "--channel", dest="channel", default="ch0:0",
                      help="Use data from the provided digital RF channel <channel>:<subchannel>.")
    parser.add_option("-l", "--length", dest="length", default=0.04, type="float",
                      help="The default data length in seconds for unframed data.")
    parser.add_option("-b", "--bins", dest="bins", default=128,
                      type="int", help="The number of time bins for the STI.")
    parser.add_option("-f", "--frames", dest="frames", default=4,
                      type="int", help="The number of sub-panel frames in the plot.")
    parser.add_option("-n", "--num_fft", dest="num_fft", default=128,
                      type="int", help="The number of FFT bints for the STI.")
    parser.add_option("-i", "--integration", dest="integration", default=1,
                      type="int", help="The number of rasters to integrate for each plot.")
    parser.add_option("-d", "--decimation", dest="decimation", default=1,
                      type="int", help="The decimation factor for the data (integer).")
    parser.add_option("-m", "--mean", dest="mean", action="store_true", default=False,
                      help="Remove the mean from the data at the PSD processing step.")
    parser.add_option("-z", "--zaxis", dest="zaxis", default=None,
                      type="string", help="zaxis colorbar setting e.g. -50:50")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose",
                      default=False, help="Print status messages to stdout.")
    parser.add_option("-o", "--outname", dest="outname", default=None,
                      type=str, help="Name of file that figure will be saved under.")
    parser.add_option("-a", "--appear", action="store_true", dest="appear",
                      default=False, help="Makes the plot appear through pyplot show.")
    if str_input is None:
        (options, args) = parser.parse_args()
    else:
        (options, args) = parser.parse_args(str_input)

    return (options, args)

#
# MAIN PROGRAM
#

# Setup Defaults
if __name__ == "__main__":
    """
        Needed to add main function to use outside functions outside of module.
    """

    # Parse the Command Line for configuration
    (options, args) = parse_command_line()

    if options.path == None:
        print "Please provide an input source with the -p option!"
        sys.exit(1)

    # Activate the DataPlotter
    dpc = DataPlotter(options)

    dpc.plot()



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""

drf_cross_sti.py
$Id$

Create a cross spectral time intensity summary plot for the given data sets.

"""


import traceback
import sys
import string
import os
import optparse
import itertools as it


import matplotlib.pyplot
import matplotlib.mlab
import matplotlib.gridspec

import scipy
import scipy.signal

import numpy
import numpy.fft
import time
import datetime
import dateutil
import pytz

import digital_rf as drf

matplotlib.rc('axes', hold=False)

class DataPlotter:
    def __init__(self, control):
        """__init__: Initializes a data plotter for STI plotting.

        """
        self.control = control

        self.dio = []
        self.dmd = []
        self.channel = []
        self.sub_channel = []
        self.bounds = []

        for idx,p in enumerate(self.control.path):

            ch = string.split(self.control.channel[idx],':')
            self.channel.append(ch[0])
            self.sub_channel.append(int(ch[1]))

            # open digital RF path
            self.dio.append(drf.DigitalRFReader(p))

            if self.control.verbose:
                print 'bounds:', self.dio[idx].get_bounds(self.channel[idx])

            self.bounds.append(self.dio[idx].get_bounds(self.channel[idx]))

            # oepn digital metadata path
            self.dmd.append(drf.DigitalMetadataReader(p + '/' + self.channel[idx] + '/metadata'))

        # processing pair list
        pl = range(len(self.dio))

        if self.control.xtype == 'self':
            self.xlist = list(it.product(pl,repeat=2))

        elif self.control.xtype == 'pairs':
            args = [iter(pl)] * 2
            self.xlist = list(it.izip_longest(*args))

        elif self.control.xtype == 'combo':
            self.xlist = list(it.combinations(pl,2))

        elif self.control.xtype == 'permute':
            self.xlist = list(it.permutations(pl,2))
        else:
            print 'unknown processing pair type ', self.control.xtype
            sys.exit(1)

        print 'pair list ', pl
        print 'xlist ', self.xlist

        # Figure setup
        # two plots coherence and phase for each pair
        self.f = []
        self.gridspec = []
        self.subplots = []

        for n in numpy.arange(len(self.xlist)):
            f = matplotlib.pyplot.figure(figsize=(7,numpy.min([numpy.max([4,self.control.frames]),7])), dpi=128)
            self.f.append(f)

            gridspec = matplotlib.gridspec.GridSpec(self.control.frames*2, 1)
            self.gridspec.append(gridspec)

            subplots = []
            self.subplots.append(subplots)

        """ Setup the subplots for this display """
        for n in numpy.arange(len(self.xlist)):
            for m in numpy.arange(self.control.frames*2):
                ax = self.f[n].add_subplot(self.gridspec[n][m])
                self.subplots[n].append(ax)


    def plot(self):
        """

            Iterate over the data set and plot the Cross STI into the subplot panels. Each
            panel is divided into a provided number of bins of a given integration
            length. Strides between the panels are made between integrations. Both coherence and
            phase are plotted.

        """

        # initialize outside the loop to avoid memory leak

        plot_a = None

        # initial plotting scales
        vmin = 0
        vmax = 0
        pmin = 0
        pmax = 0

        for fidx,xpair in enumerate(self.xlist):

            xidx, yidx = xpair

            if self.control.verbose:
                print 'pair is : ', xidx, yidx

            # sample rate
            xsr = self.dio[xidx].get_properties(self.channel[xidx])['samples_per_second']
            ysr = self.dio[yidx].get_properties(self.channel[yidx])['samples_per_second']

            if self.control.verbose:
                print 'sample rate, x: ', xsr, ' y: ', ysr

            if xsr == ysr:
                sr = xsr
            else:
                print 'problem, sample rates of data must currently match!'
                sys.exit(1)


            # initial time info
            xb = self.bounds[xidx]
            yb = self.bounds[yidx]

            if self.control.verbose:
                print 'data bounds, xb: ', xb, ' yb: ', yb

            b = (numpy.max([xb[0],yb[0]]),numpy.min([xb[1],yb[1]]))

            if self.control.start:
                dtst0 = dateutil.parser.parse(self.control.start)
                st0 = (dtst0 - datetime.datetime(1970,1,1,tzinfo=pytz.utc)).total_seconds()
                st0 = int(st0*sr)
            else:
                st0 = int(b[0])


            if self.control.end:
                dtst0 = dateutil.parser.parse(self.control.end)
                et0 = (dtst0 - datetime.datetime(1970,1,1,tzinfo=pytz.utc)).total_seconds()
                et0 = int(et0*sr)
            else:
                et0 = int(b[1])

            if self.control.verbose:

                print 'start sample st0: ', st0
                print 'end sample et0: ', et0

            blocks = self.control.bins * self.control.frames

            samples_per_stripe = self.control.num_fft * self.control.integration * self.control.decimation
            total_samples = blocks * samples_per_stripe

            if total_samples > (et0 - st0):
                print 'Insufficient samples for %d samples per stripe and %d blocks between %ld and %ld' % (samples_per_stripe, blocks, st0, et0)
                return

            stripe_stride = (et0 - st0) / blocks

            bin_stride = stripe_stride / self.control.bins

            start_sample = st0

            print 'first ', start_sample

            # get metadata
            # this could be done better to ensure we catch frequency or sample rate
            # changes
            xmdf = self.dio[xidx].read_metadata(st0, et0, self.channel[xidx])
            try:
                xmd = xmdf[xmdf.keys()[0]]
                xcfreq = xmd['center_frequencies'].ravel()[self.sub_channel[xidx]]
            except (IndexError, KeyError):
                xcfreq = 0.0
            ymdf = self.dio[yidx].read_metadata(st0, et0, self.channel[yidx])
            try:
                ymd = ymdf[ymdf.keys()[0]]
                ycfreq = ymd['center_frequencies'].ravel()[self.sub_channel[yidx]]
            except (IndexError, KeyError):
                ycfreq = 0.0
            print 'center frequencies ', xcfreq, ycfreq

            if self.control.verbose:
                print 'processing info : ', self.control.frames, self.control.bins, samples_per_stripe, bin_stride

            for p in numpy.arange(0,self.control.frames*2,2):

                sti_csd_data_coherence = numpy.zeros([self.control.num_fft, self.control.bins],numpy.float)
                sti_csd_data_phase = numpy.zeros([self.control.num_fft, self.control.bins],numpy.float)

                sti_times = numpy.zeros([self.control.bins], numpy.complex128)

                for b in numpy.arange(self.control.bins):

                    if self.control.verbose:
                        print 'read vector :', self.channel, start_sample, samples_per_stripe

                    xd_vec = self.dio[xidx].read_vector(start_sample, samples_per_stripe, self.channel[xidx])
                    xdata = xd_vec[:,self.sub_channel[xidx]]

                    yd_vec = self.dio[yidx].read_vector(start_sample, samples_per_stripe, self.channel[yidx])
                    ydata = yd_vec[:,self.sub_channel[yidx]]

                    if self.control.decimation > 1:
                        xdata = scipy.signal.decimate(xdata,self.control.decimation)
                        ydata = scipy.signal.decimate(ydata,self.control.decimation)
                        sample_freq = sr / self.control.decimation
                    else:
                        sample_freq = sr

                    if self.control.mean:
                        detrend_fn = matplotlib.mlab.detrend_mean
                    else:
                        detrend_fn = matplotlib.mlab.detrend_none

                    try:
                        csd_data, freq_axis = matplotlib.mlab.csd(xdata,ydata, NFFT=self.control.num_fft, Fs=float(sample_freq), sides='default', detrend=detrend_fn,scale_by_freq=False)
                    except:
                        traceback.print_exc(file=sys.stdout)

                    sti_csd_data_coherence[:,b] = 10.0*numpy.log10(numpy.absolute(csd_data) + 1E-12)
                    sti_csd_data_phase[:,b] = numpy.angle(csd_data)


                    sti_times[b] = start_sample / sr

                    start_sample += stripe_stride


                # Now Plot the Data
                ax = self.subplots[fidx][p]
                ax1 = self.subplots[fidx][p+1]

                # determine image x-y extent
                extent = 0, self.control.bins, -numpy.max(freq_axis)*1.1/1e3, numpy.max(freq_axis)*1.1/1e3
                # determine image color extent in log scale units

                Pss = sti_csd_data_coherence
                Pss2 = sti_csd_data_phase
                vmin = numpy.real(numpy.median(Pss) - 6.0)
                vmax = numpy.real(numpy.median(Pss) + (numpy.max(Pss) - numpy.median(Pss)) * 0.61803398875 + 50.0)
                vmin2 = -numpy.pi*1.05
                vmax2 = numpy.pi*1.05


                if self.control.zaxis:
                    vmin = int(string.split(self.control.zaxis,':')[0])
                    vmax = int(string.split(self.control.zaxis,':')[1])
                else:
                    vmin = numpy.real(numpy.median(Pss) - 6.0)
                    vmax = numpy.real(numpy.median(Pss) + (numpy.max(Pss) - numpy.median(Pss)) * 0.61803398875 + 50.0)

                im = ax.imshow(Pss, cmap='jet', origin='lower', extent=extent, interpolation='nearest', vmin=vmin, vmax=vmax,aspect='auto')
                im2 = ax1.imshow(Pss2, cmap='coolwarm', origin='lower', extent=extent, interpolation='nearest', vmin=vmin2, vmax=vmax2, aspect='auto')

                ax.set_ylabel('f (kHz)', fontsize=8)

                # plot dates

                tick_spacing = numpy.arange(self.control.bins/8, self.control.bins, self.control.bins/8)
                ax.set_xticks(tick_spacing)
                ax1.set_xticks(tick_spacing)
                tick_labels = []

                for s in tick_spacing:
                    tick_time = sti_times[s]

                    if tick_time == 0:
                        tick_string = ''
                    else:
                        gm_tick_time = time.gmtime(numpy.real(tick_time))
                        tick_string = '%02d:%02d:%02d' % (gm_tick_time[3], gm_tick_time[4], gm_tick_time[5])
                        tick_labels.append(tick_string)


                ax.set_xticklabels(tick_labels)
                ax1.set_xticklabels(tick_labels)

                # set the font sizes
                for tk in ax.get_xticklabels(): tk.set_size(8)
                for tk in ax.get_yticklabels(): tk.set_size(8)
                for tk in ax1.get_xticklabels(): tk.set_size(8)
                for tk in ax1.get_yticklabels(): tk.set_size(8)

                print 'last ', start_sample

            # create a time stamp
            start_time = st0 / sr
            srt_time = time.gmtime(start_time)
            sub_second = int(round((start_time - int(start_time)) * 100))

            timestamp = "%d-%02d-%02d %02d:%02d:%02d.%02d UT" % (srt_time[0],srt_time[1],srt_time[2],srt_time[3],srt_time[4],srt_time[5],sub_second)

            self.f[fidx].suptitle('%s %s %4.2f MHz (%d,%d)' % (self.control.title,timestamp, xcfreq/1E6, xidx, yidx) ,fontsize=8)


            #ax.legend(fontsize=8)
            ax.set_xlabel('time (UTC)', fontsize=8)
            ax1.set_xlabel('time (UTC)', fontsize=8)
            # fixup ticks

            for tk in ax.get_xticklabels(): tk.set_size(8)
            for tk in ax.get_yticklabels(): tk.set_size(8)
            for tk in ax1.get_xticklabels(): tk.set_size(8)
            for tk in ax1.get_yticklabels(): tk.set_size(8)


            self.gridspec[fidx].update()
            print "show plot"
            self.f[fidx].tight_layout()

            self.f[fidx].subplots_adjust(top=0.95,right=0.88)
            cax = self.f[fidx].add_axes([0.9, 0.55, 0.015, 0.4])
            cax1 = self.f[fidx].add_axes([0.9, 0.10, 0.015, 0.4])
            self.f[fidx].colorbar(im, cax=cax)
            self.f[fidx].colorbar(im2, cax=cax1)
        matplotlib.pyplot.show()



def parse_command_line():
    parser = optparse.OptionParser()

    parser.add_option("-t", "--title",dest="title",default='Digital RF Data', help="Use title provided for the data.")
    parser.add_option("-s", "--start",dest="start",default=None, help="Use the provided start time instead of the first time in the data. format is ISO8601: 2015-11-01T15:24:00Z")
    parser.add_option("-e", "--end",dest="end",default=None, help="Use the provided end time for the plot. format is ISO8601: 2015-11-01T15:24:00Z")

    parser.add_option("-p", "--path",dest="path",action="append",help="Use data from the provided digital RF data <paths>. More than one required.")
    parser.add_option("-c", "--channel",dest="channel",action="append",help="Use data from the provided digital RF channel <channel>:<subchannel>. More than one required.")
    parser.add_option("-l", "--length",dest="length",default=0.04,type="float",help="The default data length in seconds for unframed data.")
    parser.add_option("-b", "--bins",dest="bins",default=128,type="int",help="The number of time bins for the STI.")
    parser.add_option("-f", "--frames",dest="frames",default=4,type="int",help="The number of sub-panel frames in the plot.")
    parser.add_option("-n", "--num_fft",dest="num_fft",default=128,type="int",help="The number of FFT bints for the STI.")
    parser.add_option("-i", "--integration",dest="integration",default=1,type="int",help="The number of rasters to integrate for each plot.")
    parser.add_option("-d", "--decimation",dest="decimation",default=1,type="int",help="The decimation factor for the data (integer).")
    parser.add_option("-m", "--mean",dest="mean",action="store_true", default=False,help="Remove the mean from the data at the PSD processing step.")
    parser.add_option("-z", "--zaxis",dest="zaxis",default=None,type="string",help="zaxis colorbar setting e.g. -50:50")
    parser.add_option("-x", "--xtype",dest="xtype",default="pairs",type="string",help="Cross combination type : pairs, combo, self, permute")
    parser.add_option("-v", "--verbose",action="store_true",dest="verbose",default=False,help="Print status messages to stdout.")

    (options, args) = parser.parse_args()

    return (options, args)


#
# MAIN PROGRAM
#

# Setup Defaults
"""


"""

# Parse the Command Line for configuration
(options, args) = parse_command_line()

if options.path == None:
    print "Please provide an input source with the -p option!"
    sys.exit(1)

# Activate the DataPlotter
dpc = DataPlotter(options)

dpc.plot()



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""example of digital_metadata.read_metadata

$Id$
"""

# Millstone imports
import digital_rf
import numpy
metadata_dir = '/tmp/test_metadata'
stime = 1447082580

try:
    obj = digital_rf.DigitalMetadataReader(metadata_dir)
except:
    print('Be sure you run test_write_digital_metadata.py before running this test code.')
    raise
print('init okay')
start_idx = int(numpy.uint64(stime * obj.get_samples_per_second()))
first_sample, last_sample = obj.get_bounds()
print('bounds are %i to %i' % (first_sample, last_sample))

fields = obj.get_fields()
print('Available fields are <%s>' % (str(fields)))

print('first read - just get one column simple_complex')
data_dict = obj.read(start_idx, start_idx + 2, 'single_complex')
for key in data_dict.keys():
    print((key, data_dict[key]))

print('second read - just 2 columns: simple_complex and numpy_obj')
data_dict = obj.read(start_idx, start_idx + 2, ('single_complex', 'numpy_obj'))
for key in data_dict.keys():
    print((key, data_dict[key]))

print('third read - get all columns')
data_dict = obj.read(start_idx, start_idx + 2)
for key in data_dict.keys():
    print((key, data_dict[key]))

print('just get latest metadata')
latest_meta = obj.read_latest()
print(latest_meta)

print('test of get_samples_per_second')
sps = obj.get_samples_per_second()
print(sps)



#!python
# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""verify_digital_rf_upconvert will test if a digital rf 2 subdirectory matches sample for sample a digital rf 1 subdirectory.

$Id$
"""

# standard python imports
import sys
import argparse
import traceback

# third party imports
import numpy

# Millstone imports
import digital_rf
from digital_rf import digital_rf_deprecated_hdf5  # for reading old formatter

read_len = 1000000  # default read len


### main begins here ###
if __name__ == '__main__':

    # command line interface
    parser = argparse.ArgumentParser(
        description='verify_digital_rf_upconvert will test if a digital rf 2 subdirectory matches sample for sample a digital rf 1 subdirectory. Prints True or False')
    parser.add_argument('--drf2', metavar='drf2Dir',
                        help='Digital RF v 2 top level directory to be compared', required=True)
    parser.add_argument('--drf1', metavar='drf1Dir',
                        help='Digital RF v 1 top level directory to be compared', required=True)
    args = parser.parse_args()

    try:
        reader2 = digital_rf.DigitalRFReader(args.drf2)
    except:
        traceback.print_exc()
        sys.exit(-1)

    try:
        reader1 = digital_rf_deprecated_hdf5.read_hdf5(args.drf1)
    except:
        traceback.print_exc()
        sys.exit(-1)

    # compare each channel separately
    for channel in reader2.get_channels():
        print('comparing channel %s' % (channel))
        bounds2 = reader2.get_bounds(channel)
        bounds1 = reader1.get_bounds(channel)

        cont_blocks_2 = reader2.get_continuous_blocks(
            bounds2[0], bounds2[1], channel)
        cont_blocks_1 = reader1.get_continuous_blocks(
            bounds1[0], bounds1[1], channel)
        for i, key in enumerate(cont_blocks_2.keys()):
            s2 = key
            n2 = cont_blocks_2[key]
            s1 = cont_blocks_1[i, 0]
            n1 = cont_blocks_1[i, 1]
            if n1 > read_len:
                # keep array size reasonable
                n1 = read_len
            data1 = reader1.read_vector_raw(s1, n1, channel)
            data2 = reader2.read_vector_raw(s1, n1, channel)
            if not numpy.array_equal(data1, data2):
                print(data1)
                print(data2)
                print('False')
                sys.exit(-1)

    print('True')



# ----------------------------------------------------------------------------
# Copyright (c) 2017 Massachusetts Institute of Technology (MIT)
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
#
# The full license is in the LICENSE file, distributed with this software.
# ----------------------------------------------------------------------------
"""Read and write data/metadata in Digital Metadata HDF5 format.

It uses h5py to read and write to HDF5 files.

Reading/writing functionality is available from two classes:
DigitalMetadataReader and DigitalMetadataWriter.

"""
from __future__ import print_function

import collections
import copy
import datetime
import fractions
import glob
import itertools
import os
import re
import time
import traceback
import urllib2
from collections import defaultdict

# third party imports
import h5py
try:
    import pandas
except ImportError:
    pass
import numpy
import packaging.version
import six
from six.moves import zip

# local imports
from . import list_drf
from ._version import __version__

__all__ = (
    'DigitalMetadataReader', 'DigitalMetadataWriter',
)


def _recursive_items(d, prefix='', visited=None):
    """Generator of (key, value) pairs for a dict, recursing into sub-dicts.

    Sub-dictionary (key, value) pairs will have '[parent_key]/' prepended
    to their key name.


    Parameters
    ----------

    d : dict
        The dictionary to iterate over.

    prefix : string
        The starting prefix to be added to keys to produce the returned name.

    visited : set | None
        Set of already visited dictionary ids, to flag infinite recursion.
        If None, the empty set is used.


    Yields
    ------

    (key, val) : tuple
        Key name (with prefix and parent dictionary key added) and value pairs
        for an entry in the dictionary or a sub-dictionary.

    """
    if visited is None:
        visited = set()
    visited.add(id(d))
    for k, v in six.iteritems(d):
        name = prefix + k
        if isinstance(v, dict):
            if id(v) not in visited:
                for subk, subv in _recursive_items(v, name + '/', visited):
                    yield subk, subv
            else:
                errstr = 'Infinite loop in data - dict <%s> passed in twice.'
                raise ValueError(errstr % str(v)[0:500])
        else:
            yield name, v


class DigitalMetadataWriter:
    """Write data in Digital Metadata HDF5 format."""

    _min_version = packaging.version.parse('2.5')
    _max_version = packaging.version.parse(__version__)
    # increment when format changes are made, must be <= __version__
    _writer_version = packaging.version.parse('2.5')

    def __init__(
        self, metadata_dir, subdir_cadence_secs, file_cadence_secs,
        sample_rate_numerator, sample_rate_denominator, file_name,
    ):
        """Initialize writer to channel directory with given parameters.

        Parameters
        ----------

        metadata_dir : string
            The directory where this channel is to be written. It must already
            exist and be writable.

        subdir_cadence_secs : int
            The number of seconds of metadata to store in each subdirectory.
            The timestamp of any subdirectory will be an integer multiple of
            this value.

        file_cadence_secs : int
            The number of seconds of metadata to store in each file. Note that
            an integer number of files must exactly span a subdirectory,
            implying::

                (subdir_cadence_secs % file_cadence_secs) == 0

            This class enforces the rule that file name timestamps will be in
            the list::

                range(subdirectory_timestamp,
                      subdirectory_timestamp + subdir_cadence_secs,
                      file_cadence_secs)

        sample_rate_numerator : int
            Numerator of sample rate in Hz.

        sample_rate_denominator : int
            Denominator of sample rate in Hz.

        file_name : string
            Prefix for metadata file names. Files in each subdirectory will be
            named: "`file_name`@<timestamp>.h5".

        """
        # verify all input arguments
        if not os.access(metadata_dir, os.W_OK):
            errstr = 'metadata_dir %s does not exist or is not writable'
            raise IOError(errstr % metadata_dir)
        self._metadata_dir = metadata_dir

        if (subdir_cadence_secs != int(subdir_cadence_secs)
                or subdir_cadence_secs < 1):
            errstr = (
                'subdir_cadence_secs must be positive integer, not %s'
            )
            raise ValueError(errstr % str(subdir_cadence_secs))
        self._subdir_cadence_secs = int(subdir_cadence_secs)

        if (file_cadence_secs != int(file_cadence_secs)
                or file_cadence_secs < 1):
            errstr = (
                'file_cadence_secs must be positive integer, not %s'
            )
            raise ValueError(errstr % str(file_cadence_secs))
        self._file_cadence_secs = int(file_cadence_secs)
        if ((self._subdir_cadence_secs % self._file_cadence_secs)
                != 0):
            raise ValueError(
                '(subdir_cadence_secs % file_cadence_secs) != 0'
            )

        if not isinstance(file_name, six.string_types):
            errstr = 'file_name must be a string, not type %s'
            raise ValueError(errstr % str(type(file_name)))
        self._file_name = file_name

        if (sample_rate_numerator != int(sample_rate_numerator)
                or sample_rate_numerator < 1):
            errstr = (
                'sample_rate_numerator must be positive integer, not %s'
            )
            raise ValueError(errstr % str(sample_rate_numerator))
        self._sample_rate_numerator = int(sample_rate_numerator)

        if (sample_rate_denominator != int(sample_rate_denominator)
                or sample_rate_denominator < 1):
            errstr = (
                'sample_rate_denominator must be positive integer, not %s'
            )
            raise ValueError(errstr % str(sample_rate_denominator))
        self._sample_rate_denominator = int(sample_rate_denominator)

        # have to go to uint64 before longdouble to ensure correct conversion
        # from int
        self._samples_per_second = (
            numpy.longdouble(numpy.uint64(self._sample_rate_numerator)) /
            numpy.longdouble(numpy.uint64(self._sample_rate_denominator))
        )

        if os.access(
            os.path.join(self._metadata_dir, 'dmd_properties.h5'),
            os.R_OK,
        ) or os.access(
            os.path.join(self._metadata_dir, 'metadata.h5'),
            os.R_OK,
        ):
            self._parse_properties()
        else:
            self._digital_metadata_version = self._writer_version.base_version
            self._fields = None  # No data written yet
            self._write_properties()

    def get_samples_per_second(self):
        """Return the sample rate in Hz as a numpy.longdouble."""
        return self._samples_per_second

    def write(self, samples, data):
        """Write new metadata to the Digital Metadata channel.

        Parameters
        ----------

        samples : list | 1-D array | int | float
            A single sample index or an list of sample indices, given in
            the number of samples since the epoch (t_since_epoch*sample_rate),
            for the data to be written.

        data : list of dicts | dict
            If a list of dicts, each dictionary provides the metadata to be
            written for each corresponding sample (`data` must have the same
            length as `samples`). The dictionary keys give the field names,
            while the values must be HDF5-compatible (numpy) objects or sub-
            dictionaries meeting the same requirement.

            If a dict, the keys give the field names and each value must be
            one of the following:

                - a 1-D numpy array or list/tuple of numpy objects with length
                    equal to ``len(samples)`` giving the metadata corresponding
                    to each sample index in `samples`
                - a single value or numpy array of shape different than
                    ``(len(samples),)``, giving the metadata for *all* of the
                    samples indices in `samples`
                - another dictionary with keys that are valid Group name
                    strings and leaf values that are one of the above

            The fields should always be the same each time the write method is
            called to ensure that the fields are consistently present when
            reading.

        """
        try:
            samples = numpy.array(
                samples, dtype=numpy.uint64, copy=False, ndmin=1,
            )
        except (TypeError, ValueError):
            raise ValueError(
                'Values in `samples` must be convertible to uint64'
            )

        N = len(samples)
        if N == 0:
            raise ValueError('`samples` must not be empty')

        if isinstance(data, dict):
            if self._fields is None:
                self._set_fields(data.keys())

            keyval_iterators = []
            for key, val in _recursive_items(data):
                if not isinstance(val, six.string_types):
                    try:
                        if len(val) == N:
                            it = zip(itertools.repeat(key, N), val)
                            keyval_iterators.append(it)
                            continue
                    except TypeError:
                        pass
                # val is a string, doesn't have a length, or len(val) != N
                it = itertools.repeat((key, val), N)
                keyval_iterators.append(it)
            keyvals = zip(*keyval_iterators)
        elif len(data) == N:
            if self._fields is None:
                self._set_fields(data[0].keys())

            keyvals = (_recursive_items(d) for d in data)
        else:
            errstr = (
                '`data` must be a dict or list of dicts with length equal to'
                ' the length of `samples`.'
            )
            raise ValueError(errstr)

        return self._write(samples, keyvals)

    def _write(self, samples, keyvals):
        """Write new metadata to the Digital Metadata channel.

        This function does no input checking, see `write` for that.


        Parameters
        ----------

        samples : 1-D numpy array of type uint64 sorted in ascending order
            An array of sample indices, given in the number of samples since
            the epoch (time_since_epoch*sample_rate).

        keyvals : iterable of iterables same length as `samples`
            Each element of this iterable corresponds to a sample in `samples`
            and should be another iterable that produces (key, value) pairs to
            write for that sample.

        """
        grp_iter = self._sample_group_generator(samples)
        for grp, keyval in zip(grp_iter, keyvals):
            for key, val in keyval:
                if val is not None:
                    grp.create_dataset(key, data=val)

    def _sample_group_generator(self, samples):
        """Generator that yields HDF5 group for each sample in `samples`.

        Parameters
        ----------

        samples : 1-D numpy array of type uint64 sorted in ascending order
            An array of sample indices, given in the number of samples since
            the epoch (time_since_epoch*sample_rate).


        Yields
        ------

        grp : h5py.Group
            HDF5 group for the sample. The group is located in the appropriate
            Digital Metadata file and takes its name from the sample index.

        """
        samples_per_file = self._file_cadence_secs * self._samples_per_second
        for file_idx, sample_group in itertools.groupby(
            samples, lambda s: numpy.uint64(s / samples_per_file),
        ):
            file_ts = file_idx * self._file_cadence_secs
            file_basename = '%s@%i.h5' % (self._file_name, file_ts)

            start_sub_ts = (
                (file_ts//self._subdir_cadence_secs)*self._subdir_cadence_secs
            )
            sub_dt = datetime.datetime.utcfromtimestamp(start_sub_ts)
            subdir = os.path.join(
                self._metadata_dir, sub_dt.strftime('%Y-%m-%dT%H-%M-%S'),
            )

            if not os.path.exists(subdir):
                os.makedirs(subdir)
            this_file = os.path.join(subdir, file_basename)

            with h5py.File(this_file, 'a') as f:
                for sample in sample_group:
                    try:
                        grp = f.create_group(str(sample))
                    except ValueError:
                        errstr = (
                            'Sample %i already in data: no overwriting allowed'
                        )
                        raise IOError(errstr % sample)
                    yield grp

    def _set_fields(self, field_names):
        """Set the field names used in this metadata channel.

        This method sets both the `_fields` attribute and writes the field
        names to the channels top-level 'dmd_properties.h5' file.


        Parameters
        ----------

        field_names : list
            List of field names used in this metadata channel.

        """
        # build recarray and self._fields
        recarray = numpy.recarray(
            (len(field_names),), dtype=[('column', '|S128')],
        )
        self._fields = field_names
        self._fields.sort()  # for reproducability, use alphabetic order
        for i, key in enumerate(self._fields):
            recarray[i] = (key,)

        # write recarray to metadata
        properties_file_path = os.path.join(
            self._metadata_dir, 'dmd_properties.h5',
        )
        with h5py.File(properties_file_path, 'a') as f:
            f.create_dataset('fields', data=recarray)

    def _parse_properties(self):
        """Check writer properties against existing ones for the channel.

        When a metadata channel already exists on disk, call this method when
        creating a new DigitalMetadataWriter to check its parameters against
        the existing ones. The `fields` attribute and metadata version of the
        current writer will be set according to parameters found on disk.

        Raises
        ------

        ValueError
            If the DigitalMetadataWriter parameters do not match those on disk.

        IOError
            If the Digital Metadata version of the existing metadata on disk
            is not compatible with this software version.

        """
        # reader will raise IOError if existing properties can't be read
        # because of version incompatibilities
        org_obj = DigitalMetadataReader(self._metadata_dir, accept_empty=True)
        # but we also have to check if the metadata is the current major
        # version so we know we can continue writing to it
        self._digital_metadata_version = org_obj._digital_metadata_version
        self._check_compatible_version()
        attr_list = (
            '_subdir_cadence_secs', '_file_cadence_secs',
            '_sample_rate_numerator',  '_sample_rate_denominator',
            '_file_name',
        )
        for attr in attr_list:
            if getattr(self, attr) != getattr(org_obj, attr):
                errstr = 'Mismatched %s: %s versus %s'
                raise ValueError(errstr % (attr,
                                           getattr(self, attr),
                                           getattr(org_obj, attr)))
        self._fields = getattr(org_obj, '_fields')

    def _check_compatible_version(self):
        version = packaging.version.parse(self._digital_metadata_version)

        if (version >= self._min_version) and (version <= self._max_version):
            pass
        else:
            errstr = (
                'Existing Digital Metadata files are version %s, which is'
                ' not in the range required (%s to %s).'
            )
            raise IOError(errstr % (version.base_version,
                                    self._min_version.base_version,
                                    self._max_version.base_version))

    def _write_properties(self):
        """Write Digital Metadata properties to dmd_properties.h5 file."""
        properties_file_path = os.path.join(
            self._metadata_dir, 'dmd_properties.h5',
        )
        with h5py.File(properties_file_path, 'w') as f:
            f.attrs['subdir_cadence_secs'] = self._subdir_cadence_secs
            f.attrs['file_cadence_secs'] = self._file_cadence_secs
            f.attrs['sample_rate_numerator'] = self._sample_rate_numerator
            f.attrs['sample_rate_denominator'] = self._sample_rate_denominator
            f.attrs['file_name'] = self._file_name
            f.attrs['digital_metadata_version'] = \
                self._digital_metadata_version

    def __str__(self):
        """String summary of the DigitalMetadataWriter's parameters."""
        ret_str = ''
        attr_list = (
            '_subdir_cadence_secs', '_file_cadence_secs',
            '_samples_per_second', '_file_name',
        )
        for attr in attr_list:
            ret_str += '%s: %s\n' % (attr, str(getattr(self, attr)))
        if self._fields is None:
            ret_str += '_fields: None\n'
        else:
            ret_str += '_fields:\n'
            for key in self._fields:
                ret_str += '\t%s\n' % (key)
        return ret_str


class DigitalMetadataReader:
    """Read data in Digital Metadata HDF5 format."""

    _min_version = packaging.version.parse('2.0')

    def __init__(self, metadata_dir, accept_empty=True):
        """Initialize reader to metadata channel directory.

        Channel parameters are read from the attributes of the top-level file
        'dmd_properties.h5' in the `metadata_dir`.


        Parameters
        ----------

        metadata_dir : string
            Path to metadata channel directory, which contains a
            'dmd_properties.h5' file and timestamped subdirectories containing
            data.

        accept_empty : bool, optional
            If True, do not raise an IOError if the 'dmd_properties.h5' file is
            empty. If False, raise an IOError in that case and delete the
            empty 'dmd_properties.h5' file.


        Raises
        ------

        IOError
            If 'dmd_properties.h5' file is not found in `metadata_dir` or if
            `accept_empty` is False and the 'dmd_properties.h5' file is empty.

        """
        self._metadata_dir = metadata_dir
        if self._metadata_dir.find('http://') != -1:
            self._local = False
            # put properties file in /tmp/dmd_properties_%i.h5 % (pid)
            url = os.path.join(self._metadata_dir, 'dmd_properties.h5')
            try:
                f = urllib2.urlopen(url)
            except (urllib2.URLError, urllib2.HTTPError):
                url = os.path.join(self._metadata_dir, 'metadata.h5')
                f = urllib2.urlopen(url)
            tmp_file = os.path.join(
                '/tmp', 'dmd_properties_%i.h5' % (os.getpid()),
            )
            fo = open(tmp_file, 'w')
            fo.write(f.read())
            f.close()
            fo.close()

        else:
            self._local = True
            # list and match first properties file
            tmp_file = next(
                (f for f in sorted(glob.glob(os.path.join(
                    metadata_dir, list_drf.GLOB_DMDPROPFILE,
                ))) if re.match(list_drf.RE_DMDPROP, f)),
                None,
            )
            if tmp_file is None:
                raise IOError('dmd_properties.h5 not found')

        with h5py.File(tmp_file, 'r') as f:
            try:
                subdir_cadence = f.attrs['subdir_cadence_secs'].item()
                file_cadence = f.attrs['file_cadence_secs'].item()
            except KeyError:
                # maybe an older version with subdirectory_cadence_seconds
                # and file_cadence_seconds
                subdir_cadence = f.attrs['subdirectory_cadence_seconds'].item()
                file_cadence = f.attrs['file_cadence_seconds'].item()
            self._subdir_cadence_secs = subdir_cadence
            self._file_cadence_secs = file_cadence
            try:
                try:
                    spsn = f.attrs['sample_rate_numerator'].item()
                    spsd = f.attrs['sample_rate_denominator'].item()
                except KeyError:
                    # maybe an older version with samples_per_second_*
                    spsn = f.attrs['samples_per_second_numerator'].item()
                    spsd = f.attrs['samples_per_second_denominator'].item()
            except KeyError:
                # must have an older version with samples_per_second attribute
                sps = f.attrs['samples_per_second'].item()
                spsfrac = fractions.Fraction(sps).limit_denominator()
                self._samples_per_second = numpy.longdouble(sps)
                self._sample_rate_numerator = int(spsfrac.numerator)
                self._sample_rate_denominator = int(spsfrac.denominator)
            else:
                self._sample_rate_numerator = spsn
                self._sample_rate_denominator = spsd
                # have to go to uint64 before longdouble to ensure correct
                # conversion from int
                self._samples_per_second = (
                    numpy.longdouble(numpy.uint64(
                        self._sample_rate_numerator
                    )) /
                    numpy.longdouble(numpy.uint64(
                        self._sample_rate_denominator
                    ))
                )
            self._file_name = f.attrs['file_name']
            try:
                version = f.attrs['digital_metadata_version']
            except KeyError:
                # version is before 2.3 when attribute was added
                version = '2.0'
            self._digital_metadata_version = version
            self._check_compatible_version()
            try:
                fields_dataset = f['fields']
            except KeyError:
                if not accept_empty:
                    os.remove(tmp_file)
                    errstr = (
                        'No metadata yet written to %s, removing empty'
                        ' "dmd_properties.h5"'
                    )
                    raise IOError(errstr % self._metadata_dir)
                else:
                    self._fields = None
                    return
            self._fields = []
            for i in range(len(fields_dataset)):
                self._fields.append(fields_dataset[i]['column'])

        if not self._local:
            os.remove(tmp_file)

    def get_bounds(self):
        """Get indices of first- and last-known sample as a tuple.

        Returns
        -------

        first_sample_index : int
            Index of the first sample, given in the number of samples since the
            epoch (time_since_epoch*sample_rate).

        last_sample_index : int
            Index of the last sample, given in the number of samples since the
            epoch (time_since_epoch*sample_rate).


        Raises
        ------

        IOError
            If no data or first and last sample could not be determined.

        """
        # loop through files in order to get first sample
        first_sample = None
        for path in list_drf.ilsdrf(
            self._metadata_dir, recursive=False, reverse=False,
            include_dmd=True, include_drf=False, include_dmd_properties=False,
        ):
            try:
                with h5py.File(path, 'r') as f:
                    groups = f.keys()
                    groups.sort()
                    first_sample = int(groups[0])
            except (IOError, IndexError):
                errstr = (
                    'Corrupt or empty file %s found and ignored.'
                    ' Deleting it will speed up get_bounds().'
                )
                print(errstr % path)
            else:
                break
        if first_sample is None:
            raise IOError('All attempts to read first sample failed')

        # loop through files in reverse order to get last sample
        last_sample = None
        for path in list_drf.ilsdrf(
            self._metadata_dir, recursive=False, reverse=True,
            include_dmd=True, include_drf=False, include_dmd_properties=False,
        ):
            try:
                with h5py.File(path, 'r') as f:
                    groups = f.keys()
                    groups.sort()
                    last_sample = int(groups[-1])
            except (IOError, IndexError):
                errstr = (
                    'Corrupt or empty file %s found and ignored.'
                    ' Deleting it will speed up get_bounds().'
                )
                print(errstr % path)
            else:
                break
        if last_sample is None:
            raise IOError('All attempts to read last sample failed')

        return (first_sample, last_sample)

    def get_fields(self):
        """Return list of the field names in this metadata."""
        # _fields is an internal data structure, so make a copy for the user
        return copy.deepcopy(self._fields)

    def get_sample_rate_numerator(self):
        """Return the numerator of the sample rate in Hz."""
        return self._sample_rate_numerator

    def get_sample_rate_denominator(self):
        """Return the denominator of the sample rate in Hz."""
        return self._sample_rate_denominator

    def get_samples_per_second(self):
        """Return the sample rate in Hz as a numpy.longdouble."""
        return self._samples_per_second

    def get_subdir_cadence_secs(self):
        """Return the number of seconds of data stored in each subdirectory."""
        return self._subdir_cadence_secs

    def get_file_cadence_secs(self):
        """Return the number of seconds of data stored in each file."""
        return self._file_cadence_secs

    def get_file_name_prefix(self):
        """Return the metadata file name prefix."""
        return self._file_name

    def read(
        self, start_sample=None, end_sample=None, columns=None, method=None,
    ):
        """Read metadata between start and end samples.

        Parameters
        ----------

        start_sample : None | int
            Sample index for start of read, given in the number of samples
            since the epoch (time_since_epoch*sample_rate). If None,
            `get_bounds` is called and the last sample is used.

        end_sample : None | int
            Sample index for end of read (inclusive), given in the number of
            samples since the epoch (time_since_epoch*sample_rate). If None,
            use `end_sample` equal to `start_sample`.

        columns : None | string | list of strings
            A string or list of strings giving the field/column name of
            metadata to return. If None, all available columns will be read.
            Using a string results in a different returned object than a one-
            element list containing the string, see below.

        method : None | 'pad'/'ffill'
            If None, return only samples within the given range. If 'pad' or
            'ffill', the first sample no later than `start_sample` (if any)
            will also be included so that values are forward filled into the
            desired range.


        Returns
        -------

        OrderedDict
            The dictionary's keys are the sample index for each sample of
            metadata found between `start_sample` and `end_sample` (inclusive).
            Each value is a metadata sample, given as either the column value
            (if `columns` is a string) or a dictionary with column names as
            keys and numpy objects as leaf values (if `columns` is None or
            a list).


        See Also
        --------

        read_dataframe : Read metadata into a DataFrame.
        read_flatdict : Read metadata into a flat dictionary, keyed by field.

        """
        if start_sample is None:
            _, start_sample = self.get_bounds()
        if end_sample is None:
            end_sample = start_sample
        elif start_sample > end_sample:
            errstr = 'Start sample %i more than end sample %i'
            raise ValueError(errstr % (start_sample, end_sample))

        ret_dict = collections.OrderedDict()

        if method in ('pad', 'ffill'):
            # simple forward fill until something better is needed:
            #  get start bound of data and search for metadata within
            #  [start_bound, start_sample] until last sample is found
            ffill_dict = collections.OrderedDict()
            start_bound, end_bound = self.get_bounds()
            file_list = self._get_file_list(start_bound, start_sample)
            # go through files in reverse to break at last found sample
            for this_file in reversed(file_list):
                self._add_metadata(
                    ffill_dict, this_file, columns, start_bound, start_sample,
                    is_edge=False,
                )
                if ffill_dict:
                    # get last entry of ffill_dict which will be latest found
                    # sample in the file
                    key = next(reversed(ffill_dict))
                    ret_dict[key] = ffill_dict[key]
                    break
            # increment start sample so we don't re-add any data at that sample
            start_sample += 1

        file_list = self._get_file_list(start_sample, end_sample)
        for this_file in file_list:
            if this_file in (file_list[0], file_list[-1]):
                is_edge = True
            else:
                is_edge = False
            self._add_metadata(
                ret_dict, this_file, columns, start_sample, end_sample,
                is_edge,
            )

        return ret_dict

    def read_dataframe(
        self, start_sample=None, end_sample=None, columns=None, method=None,
    ):
        """Read metadata between start and end samples into a pandas DataFrame.

        Parameters
        ----------

        start_sample : int
            Sample index for start of read, given in the number of samples
            since the epoch (time_since_epoch*sample_rate). If None,
            `get_bounds` is called and the last sample is used.

        end_sample : None | int
            Sample index for end of read (inclusive), given in the number of
            samples since the epoch (time_since_epoch*sample_rate). If None,
            use `end_sample` equal to `start_sample`.

        columns : None | string | list of strings
            A string or list of strings giving the field/column name of
            metadata to return. If None, all available columns will be read.

        method : None | 'pad'/'ffill'
            If None, return only samples within the given range. If 'pad' or
            'ffill', the first sample no later than `start_sample` (if any)
            will also be included so that values are forward filled into the
            desired range.


        Returns
        -------

        DataFrame
            Pandas DataFrame with rows corresponding to the sample index and
            columns corresponding to the metadata key.


        See Also
        --------

        read : Read metadata into an OrderedDict, keyed by sample index.
        read_flatdict : Read metadata into a flat dictionary, keyed by field.

        """
        if isinstance(columns, six.string_types):
            # preserve column name in returned dictionary so it appears in DF
            columns = [columns]
        res = self.read(
            start_sample=start_sample, end_sample=end_sample,
            columns=columns, method=method,
        )
        data = list(dict(_recursive_items(d)) for d in res.values())
        index = list(res.keys())
        return pandas.DataFrame(data, index=index)

    def read_flatdict(
        self, start_sample=None, end_sample=None, columns=None, method=None,
        squeeze=True,
    ):
        """Read metadata between start and end samples into a flat dictionary.

        Parameters
        ----------

        start_sample : int
            Sample index for start of read, given in the number of samples
            since the epoch (time_since_epoch*sample_rate). If None,
            `get_bounds` is called and the last sample is used.

        end_sample : None | int
            Sample index for end of read (inclusive), given in the number of
            samples since the epoch (time_since_epoch*sample_rate). If None,
            use `end_sample` equal to `start_sample`.

        columns : None | string | list of strings
            A string or list of strings giving the field/column name of
            metadata to return. If None, all available columns will be read.

        method : None | 'pad'/'ffill'
            If None, return only samples within the given range. If 'pad' or
            'ffill', the first sample no later than `start_sample` (if any)
            will also be included so that values are forward filled into the
            desired range.

        squeeze : bool
            If True and end_sample is None (returning a single sample), return
            the column values for the sample directly instead of as arrays with
            a first dimension of one. Additionally, if only a single column
            name is given and the result contains no subfields, return the
            value of that column instead of a dictionary.


        Returns
        -------

        dict or object
            Dictionary with keys corresponding to the fields/columns of the
            requested metadata. The values are arrays with length equal to
            the number of samples. The dictionary also has an 'index' entry
            containing an array of the sample indices. If `squeeze` is True
            and a dictionary with a single sample and non-index column would be
            returned, the non-index value is returned instead.


        See Also
        --------

        read : Read metadata into an OrderedDict, keyed by sample index.
        read_dataframe : Read metadata into a DataFrame.

        """
        if isinstance(columns, six.string_types):
            # preserve column name in returned dictionary so it appears in DF
            columns = [columns]
        res = self.read(
            start_sample=start_sample, end_sample=end_sample,
            columns=columns, method=method,
        )
        dict_of_lists = defaultdict(lambda: [numpy.nan]*len(res))
        dict_of_lists[u'index'] = res.keys()
        for k, sample_dict in enumerate(res.values()):
            for key, val in _recursive_items(sample_dict):
                dict_of_lists[key][k] = val
        if squeeze and (end_sample is None):
            flatdict = {k: v[0] for k, v in dict_of_lists.items()}
            if len(flatdict) == 2:
                del flatdict['index']
                return flatdict.popitem()[1]
            else:
                return flatdict
        else:
            return {k: numpy.array(v) for k, v in dict_of_lists.items()}

    def read_latest(self, columns=None):
        """Read the most recent metadata sample.

        This method calls `get_bounds` to find the last sample index and `read`
        to read the latest metadata at or before the last sample.

        Parameters
        ----------

        columns : None | string | list of strings
            A string or list of strings giving the field/column name of
            metadata to return. If None, all available columns will be read.
            Using a string results in a different returned object than a one-
            element list containing the string, see below.


        Returns
        -------

        dict
            Dictionary containing the latest metadata, where the key is the
            sample index and the value is the metadata sample given as either
            the column value (if `columns` is a string) or a dictionary with
            column names as keys and numpy objects as leaf values (if `columns`
            is None or a list).

        """
        start_sample, last_sample = self.get_bounds()
        return self.read(last_sample, columns=columns, method='ffill')

    def _get_file_list(self, sample0, sample1):
        """Get an ordered list of data file names that could contain data.

        This takes a first and last sample and generates the possible filenames
        spanning that time according to the subdirectory and file cadences.


        Parameters
        ----------

        sample0 : int
            Sample index for start of read, given in the number of samples
            since the epoch (time_since_epoch*sample_rate).

        sample1 : int
            Sample index for end of read (inclusive), given in the number of
            samples since the epoch (time_since_epoch*sample_rate).


        Returns
        -------

        list
            List of file paths that exist on disk, fall in the given time
            interval, and conform to the subdirectory and file cadence naming
            scheme.

        """
        # need to go through numpy uint64 to prevent conversion to float
        start_ts = int(numpy.uint64(sample0/self._samples_per_second))
        end_ts = int(numpy.uint64(sample1/self._samples_per_second))

        # convert ts to be divisible by self._file_cadence_secs
        start_ts = \
            (start_ts // self._file_cadence_secs) * self._file_cadence_secs
        end_ts = (end_ts // self._file_cadence_secs) * self._file_cadence_secs

        # get subdirectory start and end ts
        start_sub_ts = \
            (start_ts // self._subdir_cadence_secs) * self._subdir_cadence_secs
        end_sub_ts = \
            (end_ts // self._subdir_cadence_secs) * self._subdir_cadence_secs

        ret_list = []  # ordered list of full file paths to return

        for sub_ts in range(
            start_sub_ts,
            end_sub_ts + self._subdir_cadence_secs,
            self._subdir_cadence_secs,
        ):
            sub_datetime = datetime.datetime.utcfromtimestamp(sub_ts)
            subdir = sub_datetime.strftime('%Y-%m-%dT%H-%M-%S')
            # create numpy array of all file TS in subdir
            file_ts_in_subdir = numpy.arange(
                sub_ts,
                sub_ts + self._subdir_cadence_secs,
                self._file_cadence_secs,
            )
            # file has valid samples if last time in file is after start time
            # and first time in file is before end time
            valid_in_subdir = numpy.logical_and(
                file_ts_in_subdir + self._file_cadence_secs - 1 >= start_ts,
                file_ts_in_subdir <= end_ts,
            )
            valid_file_ts_list = numpy.compress(
                valid_in_subdir,
                file_ts_in_subdir,
            )
            for valid_file_ts in valid_file_ts_list:
                file_basename = '%s@%i.h5' % (self._file_name, valid_file_ts)
                full_file = os.path.join(
                    self._metadata_dir, subdir, file_basename,
                )
                # verify exists
                if not os.access(full_file, os.R_OK):
                    continue
                ret_list.append(full_file)

        return ret_list

    def _add_metadata(
        self, ret_dict, this_file, columns, sample0, sample1, is_edge,
    ):
        """Read metadata from a single file and add it to `ret_dict`.

        Parameters
        ----------

        ret_dict : OrderedDict
            Dictionary to which metadata will be added.

        this_file : string
            Full path to the file from which metadata will be read.

        columns : None | string | list of strings
            A string or list of strings giving the field/column name of
            metadata to return. If None, all available columns will be read.

        sample0 : int
            Sample index for start of read, given in the number of samples
            since the epoch (time_since_epoch*sample_rate).

        sample1 : int
            Sample index for end of read (inclusive), given in the number of
            samples since the epoch (time_since_epoch*sample_rate).

        is_edge : bool
            If True, then this is the first or last file in a sequence spanning
            indices from `sample0` to `sample1` and those read boundaries must
            be taken into account. If False, all samples from the file will be
            read ignoring `sample0` and `sample1`.

        """
        try:
            with h5py.File(this_file, 'r') as f:
                # get sorted numpy array of all samples in file
                keys = f.keys()
                idxs = numpy.fromiter(keys, numpy.int64, count=len(keys))
                idxs.sort()
                if is_edge:
                    # calculate valid samples based on sample range
                    valid = numpy.logical_and(idxs >= sample0, idxs <= sample1)
                    idxs = idxs[valid]
                for idx in idxs:
                    value = f[str(idx)]
                    if columns is None:
                        self._populate_data(ret_dict, value, idx)
                    elif isinstance(columns, six.string_types):
                        self._populate_data(ret_dict, value[columns], idx)
                    else:
                        ret_dict[idx] = {}
                        for column in columns:
                            self._populate_data(
                                ret_dict[idx], value[column], column,
                            )
        except IOError:
            # decide whether this file is corrupt, or too new, or just missing
            if os.access(this_file, os.R_OK) and os.access(this_file, os.W_OK):
                if (time.time() - os.path.getmtime(this_file) >
                        self._file_cadence_secs):
                    traceback.print_exc()
                    errstr = (
                        'WARNING: %s being deleted because it raised an error'
                        ' and is not new'
                    )
                    print(errstr % this_file)
                    os.remove(this_file)

    def _populate_data(self, ret_dict, obj, name):
        """Read data recursively from an HDF5 value and add it to `ret_dict`.

        If `obj` is a dataset, it is added to `ret_dict`. If `obj` is a group,
        a sub-dictionary is created in `ret_dict` for `obj` and populated
        recursively by calling this function on all of  the items in the `obj`
        group.

        Parameters
        ----------

        ret_dict : OrderedDict
            Dictionary to which metadata will be added.

        obj : h5py.Dataset | h5py.Group
            HDF5 value from which to read metadata.

        name : valid dictionary key
            Dictionary key in `ret_dict` under which to store the data from
            `obj`.

        """
        if isinstance(obj, h5py.Dataset):
            # [()] casts a Dataset as a numpy array
            ret_dict[name] = obj[()]
        else:
            # create a dictionary for this group
            ret_dict[name] = {}
            for key, value in obj.items():
                self._populate_data(ret_dict[name], value, key)

    def _check_compatible_version(self):
        version = packaging.version.parse(self._digital_metadata_version)

        if version >= self._min_version:
            pass
        else:
            errstr = (
                'Digital Metadata files being read are version {0}, which is'
                ' less than the required version ({1}).'
            ).format(version.base_version, self._min_version.base_version)
            raise IOError(errstr)

    def __str__(self):
        """String summary of the DigitalMetadataReader's parameters."""
        ret_str = ''
        attr_list = (
            '_subdir_cadence_secs', '_file_cadence_secs',
            '_samples_per_second', '_file_name',
        )
        for attr in attr_list:
            ret_str += '%s: %s\n' % (attr, str(getattr(self, attr)))
        if self._fields is None:
            ret_str += '_fields: None\n'
        else:
            ret_str += '_fields:\n'
            for key in self._fields:
                ret_str += '\t%s\n' % (key)
        return ret_str
